import asyncio
import difflib
import importlib
import logging
import os
import re
import sys
import textwrap
import traceback
import warnings
from collections.abc import Sequence
from typing import Annotated, Optional, Union

import click
import coloredlogs
import typer
import typer.core
from packaging.version import Version
from typer.core import TyperGroup
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import TUNNEL_ENV_VAR, isatty, set_color_flag, set_verbosity
from pymobiledevice3.exceptions import (
    AccessDeniedError,
    AfcException,
    AfcFileNotFoundError,
    CloudConfigurationAlreadyPresentError,
    ConnectionFailedError,
    ConnectionFailedToUsbmuxdError,
    ConnectionTerminatedError,
    DeprecationError,
    DeveloperModeError,
    DeveloperModeIsNotEnabledError,
    DeviceHasPasscodeSetError,
    DeviceNotFoundError,
    FeatureNotSupportedError,
    InternalError,
    InvalidServiceError,
    MessageNotSupportedError,
    MissingValueError,
    NoDeviceConnectedError,
    NotEnoughDiskSpaceError,
    NotPairedError,
    OSNotSupportedError,
    PairingDialogResponsePendingError,
    PasswordRequiredError,
    QuicProtocolNotSupportedError,
    RSDRequiredError,
    SetProhibitedError,
    StartServiceError,
    TunneldConnectionError,
    UserDeniedPairingError,
)
from pymobiledevice3.lockdown import create_using_usbmux, retry_create_using_usbmux
from pymobiledevice3.osu.os_utils import get_os_utils

coloredlogs.install(level=logging.INFO)

logging.getLogger("quic").disabled = True
logging.getLogger("asyncio").disabled = True
logging.getLogger("parso.cache").disabled = True
logging.getLogger("parso.cache.pickle").disabled = True
logging.getLogger("parso.python.diff").disabled = True
logging.getLogger("humanfriendly.prompts").disabled = True
logging.getLogger("blib2to3.pgen2.driver").disabled = True
logging.getLogger("urllib3.connectionpool").disabled = True

logger = logging.getLogger(__name__)

# For issue https://github.com/doronz88/pymobiledevice3/issues/1217, details: https://bugs.python.org/issue37373
if sys.platform == "win32":
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

INVALID_SERVICE_MESSAGE = """Failed to start service. Possible reasons are:
- If you were trying to access a developer service (developer subcommand):
    - If your device iOS version >= 15.0:
        - Make sure you first enabled "Developer Mode" via:
          > python3 -m pymobiledevice3 amfi enable-developer-mode

    - Make sure the DeveloperDiskImage/PersonalizedImage is mounted via:
      > python3 -m pymobiledevice3 mounter auto-mount

    - If your device iOS version >= 17.0:
        - Make sure you passed the --rsd option to the subcommand
          https://github.com/doronz88/pymobiledevice3#working-with-developer-tools-ios--170

- Apple removed this service, or your iOS version does not support it.

- A bug. Please file a bug report:
  https://github.com/doronz88/pymobiledevice3/issues/new?assignees=&labels=&projects=&template=bug_report.md&title=
"""

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"], "max_content_width": 400}

# Mapping of index options to import file names
CLI_GROUPS = {
    "activation": "activation",
    "afc": "afc",
    "amfi": "amfi",
    "apps": "apps",
    "backup2": "backup",
    "bonjour": "bonjour",
    "companion": "companion_proxy",
    "crash": "crash",
    "developer": "developer",
    "diagnostics": "diagnostics",
    "lockdown": "lockdown",
    "mounter": "mounter",
    "notification": "notification",
    "pcap": "pcap",
    "power-assertion": "power_assertion",
    "processes": "processes",
    "profile": "profile",
    "provision": "provision",
    "remote": "remote",
    "restore": "restore",
    "springboard": "springboard",
    "syslog": "syslog",
    "usbmux": "usbmux",
    "webinspector": "webinspector",
    "idam": "idam",
    "version": "version",
}

# Set if used the `--reconnect` option
RECONNECT = False


class Pmd3TyperGroup(TyperGroup):
    def list_commands(self, ctx: click.Context) -> list[str]:
        # Order is preserved by dict insertion; adjust if you want alphabetical
        return list(CLI_GROUPS.keys())

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command:
        if cmd_name not in CLI_GROUPS:
            self.handle_invalid_command(ctx, cmd_name)
        return self.import_and_get_command(ctx, cmd_name)

    def handle_invalid_command(self, ctx, name: str) -> None:
        suggested_commands = self.search_commands(name)
        suggestion = self.format_suggestions(suggested_commands)
        # ctx.fail raises a ClickException underneath, which Typer displays nicely
        ctx.fail(f"No such command {name!r}{suggestion}")

    @staticmethod
    def format_suggestions(suggestions: list[str]) -> str:
        if not suggestions:
            return ""
        cmds = textwrap.indent("\n".join(suggestions), " " * 4)
        return f"\nDid you mean:\n{cmds}"

    @staticmethod
    def import_and_get_command(ctx: click.Context, name: str) -> click.Command:
        module_name = f"pymobiledevice3.cli.{CLI_GROUPS[name]}"
        mod = importlib.import_module(module_name)
        # submodules expose a Typer Group named "cli"
        cli: typer.Typer = mod.cli
        return typer.main.get_command(cli)

    @staticmethod
    def highlight_keyword(text: str, keyword: str) -> str:
        return re.sub(f"({keyword})", typer.style("\\1", bold=True), text, flags=re.IGNORECASE)

    @staticmethod
    def collect_commands(command: Union[TyperGroup, click.Command]) -> Union[str, list[str]]:
        if isinstance(command, TyperGroup):  # group
            cmds = []
            for v in command.commands.values():
                child = Pmd3TyperGroup.collect_commands(v)
                if isinstance(child, list):
                    cmds.extend([f"{command.name} {c}" for c in child])
                else:
                    cmds.append(f"{command.name} {child}")
            return cmds
        return command.name or ""

    @staticmethod
    def search_commands(pattern: str) -> list[str]:
        all_commands = Pmd3TyperGroup.load_all_commands()
        matched = sorted(filter(lambda cmd: re.search(pattern, cmd), all_commands))
        if not matched:
            matched = difflib.get_close_matches(pattern, all_commands, n=20, cutoff=0.4)
        if isatty():
            matched = [Pmd3TyperGroup.highlight_keyword(cmd, pattern) for cmd in matched]
        return matched

    @staticmethod
    def load_all_commands() -> list[str]:
        all_commands: list[str] = []
        for key in CLI_GROUPS:
            module_name = f"pymobiledevice3.cli.{CLI_GROUPS[key]}"
            mod = importlib.import_module(module_name)
            if isinstance(mod.cli, typer.Typer):
                cmd = Pmd3TyperGroup.collect_commands(typer.main.get_group(mod.cli))
            else:
                cmd = Pmd3TyperGroup.collect_commands(mod.cli.commands[key])
            if isinstance(cmd, list):
                all_commands.extend(cmd)
            else:
                all_commands.append(cmd)
        return all_commands

    def resolve_command(
        self, ctx: click.Context, args: list[str]
    ) -> tuple[Optional[str], Optional[click.Command], list[str]]:
        return super().resolve_command(ctx, args)


app = InjectingTyper(
    cls=Pmd3TyperGroup,
    context_settings=CONTEXT_SETTINGS,
    no_args_is_help=True,
    # add_completion=False,
    rich_markup_mode="markdown",
    help=(
        "Swiss-army CLI for pairing, inspecting, backing up, and automating iOS devices.\n\n"
        "Docs and examples: https://github.com/doronz88/pymobiledevice3"
    ),
)


@app.callback()
def _root(
    reconnect: Annotated[
        bool,
        typer.Option(
            "--reconnect",
            help="Automatically reconnect if the device disconnects mid-command.",
            show_default=False,
        ),
    ] = False,
    verbosity: Annotated[
        int,
        typer.Option(
            "--verbose",
            "-v",
            count=True,
            help="Increase logging verbosity (repeat for more detail).",
        ),
    ] = 0,
    color: Annotated[
        bool,
        typer.Option(help="Colorize output; disable with --no-color for plain logs."),
    ] = True,
) -> None:
    """
    Top-level options for pymobiledevice3.
    """
    global RECONNECT
    RECONNECT = reconnect
    set_verbosity(verbosity)
    set_color_flag(color)


def device_might_need_tunneld(identifier: str) -> bool:
    """
    Determines if the device might require tunneling based on its product version.

    This function uses the `create_using_usbmux` context manager to establish a lockdown
    session with the specified identifier. It retrieves the device's product version,
    and checks if it is greater than or equal to version "17.0". If so, the function
    returns True, indicating that the device might require tunneling. Otherwise, it
    returns False.

    :param identifier: A string representing the device identifier.
    :return: A boolean indicating whether the device might require tunneling.
    """

    async def _device_might_need_tunneld() -> bool:
        async with await create_using_usbmux(serial=identifier) as lockdown:
            return Version(lockdown.product_version) >= Version("17.0")

    return asyncio.run(_device_might_need_tunneld())


class PossiblyMisplacedOption(click.NoSuchOption):
    def __init__(
        self,
        option_name: str,
        message: Optional[str] = None,
        possibilities: Optional[Sequence[str]] = None,
        ctx: Optional[click.Context] = None,
        suggested_ctx: Optional[click.Context] = None,
    ) -> None:
        super().__init__(option_name, message, possibilities, ctx)
        if suggested_ctx is not None:
            if ctx is not None:
                self.message += f" for subcommand: {ctx.command_path}"

            suggestion = f"{suggested_ctx.command_path} {option_name}"
            suggestion += ctx.command_path.removeprefix(suggested_ctx.command_path) if ctx is not None else " ..."

            self.message += f"\nDid you mean: {suggestion}?"

    @staticmethod
    def from_no_such_option(e: click.NoSuchOption) -> "PossiblyMisplacedOption":
        ctx = e.ctx
        while ctx:
            for param in ctx.command.params:
                if isinstance(param, typer.core.TyperOption) and (
                    e.option_name in param.opts or e.option_name in param.secondary_opts
                ):
                    break
            else:
                ctx = ctx.parent
                continue
            break

        return PossiblyMisplacedOption(e.option_name, e.message, e.possibilities, e.ctx, ctx)


def invoke_cli_with_error_handling() -> bool:
    """
    Invoke the command line interface and return `True` if the failure reason of the command was that the device was
    disconnected.
    """
    try:
        # Typer apps are callable; this executes the CLI with current sys.argv
        try:
            app(standalone_mode=False)
        except click.NoSuchOption as e:
            raise PossiblyMisplacedOption.from_no_such_option(e) from e
    except NoDeviceConnectedError:
        logger.error("Device is not connected")
        return True
    except ConnectionTerminatedError:
        logger.error("Connection was terminated abruptly")
        return True
    except NotPairedError:
        logger.error("Device is not paired")
    except UserDeniedPairingError:
        logger.error("User refused to trust this computer")
    except PairingDialogResponsePendingError:
        logger.error("Waiting for user dialog approval")
    except SetProhibitedError:
        logger.error("lockdownd denied the access")
    except MissingValueError:
        logger.error("No such value")
    except DeviceHasPasscodeSetError:
        logger.error("Cannot enable developer-mode when passcode is set")
    except DeveloperModeError:
        logger.error("Failed to enable developer-mode.")
    except ConnectionFailedToUsbmuxdError:
        logger.error("Failed to connect to usbmuxd socket. Make sure it's running.")
    except ConnectionFailedError:
        logger.error("Failed to connect to service port.")
        return True
    except MessageNotSupportedError:
        logger.error("Message not supported for this iOS version")
        traceback.print_exc()
    except InternalError:
        logger.error("Internal Error")
    except DeveloperModeIsNotEnabledError:
        logger.error(
            "Developer Mode is disabled. You can try to enable it using: "
            "python3 -m pymobiledevice3 amfi enable-developer-mode"
        )
    except (InvalidServiceError, RSDRequiredError) as e:
        should_retry_over_tunneld = False
        if isinstance(e, RSDRequiredError):
            logger.warning("Trying again over tunneld since RSD is required for this command")
            should_retry_over_tunneld = True
        elif (
            (e.identifier is not None)
            and ("developer" in sys.argv)
            and ("--tunnel" not in sys.argv)
            and device_might_need_tunneld(e.identifier)
        ):
            logger.warning("Got an InvalidServiceError. Trying again over tunneld since it is a developer command")
            should_retry_over_tunneld = True
        if should_retry_over_tunneld:
            # use a single space because Typer/Click will ignore envvars of empty strings
            os.environ[TUNNEL_ENV_VAR] = e.identifier or " "
            main()
            return False
        logger.error(INVALID_SERVICE_MESSAGE)
    except PasswordRequiredError:
        logger.error("Device is password protected. Please unlock and retry")
    except AccessDeniedError:
        logger.error(get_os_utils().access_denied_error)
    except BrokenPipeError:
        traceback.print_exc()
    except TunneldConnectionError:
        logger.error(
            "Unable to connect to Tunneld. You can start one using:\nsudo python3 -m pymobiledevice3 remote tunneld"
        )
    except DeviceNotFoundError as e:
        logger.error(f"Device not found: {e.udid}")
    except NotEnoughDiskSpaceError:
        logger.error("Not enough disk space")
    except DeprecationError:
        logger.error("failed to query MobileGestalt, MobileGestalt deprecated (iOS >= 17.4).")
    except OSNotSupportedError as e:
        logger.error(
            f"Unsupported OS - {e.os_name}. To add support, consider contributing at "
            f"https://github.com/doronz88/pymobiledevice3."
        )
    except CloudConfigurationAlreadyPresentError:
        logger.error(
            "A cloud configuration is already present on device. You must first erase the device in order "
            "to install new one:\n"
            "> pymobiledevice3 profile erase-device"
        )
    except FeatureNotSupportedError as e:
        logger.error(
            f"Missing implementation of `{e.feature}` on `{e.os_name}`. To add support, consider contributing at "
            f"https://github.com/doronz88/pymobiledevice3."
        )
    except QuicProtocolNotSupportedError:
        logger.error("Encountered a QUIC protocol error.")
    except StartServiceError as e:
        if e.message == "ServiceProhibited" and e.service_name == "com.apple.pcapd.shim.remote":
            logger.error(
                f"The {e.service_name} service is USB only (at least for some iOS versions).\n"
                "Full discussion is available in: https://github.com/doronz88/pymobiledevice3/issues/1515"
            )
        else:
            logger.error(f"Failed to start: {e.service_name} with. Received error: {e.message}.")
    except AfcFileNotFoundError as e:
        logger.error(f"File [{e.filename}] not found during afc operation: {e}")
    except AfcException as e:
        logger.error(f"Failed to perform Afc operation: {e}")
    except click.ClickException as e:
        from typer import rich_utils

        rich_utils.rich_format_error(e)
        sys.exit(e.exit_code)

    return False


def main() -> None:
    # Retry to invoke the CLI
    while invoke_cli_with_error_handling():
        # If reached here, this means the failure reason was that the device is disconnected
        if not RECONNECT:
            # If not invoked with the `--reconnect` option, break here
            break
        try:
            # Wait for the device to be available again
            lockdown = asyncio.run(retry_create_using_usbmux(None))
            asyncio.run(lockdown.close())
        except KeyboardInterrupt:
            print("Aborted.")
            break


if __name__ == "__main__":
    main()



"""Minimal WebDriverAgent (WDA) client.

This module provides:
- A simple HTTP-based client for direct WDA usage.
- An async client that routes requests through a LockdownServiceProvider
  connection (usbmux, TCP, or RSD-backed) without requiring a local forwarder.
"""

import base64
import json
from dataclasses import dataclass
from typing import Any, Optional

import requests

from pymobiledevice3.exceptions import WdaError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.service_connection import ServiceConnection

DEFAULT_WDA_PORT = 8100
DEFAULT_WDA_URL = "http://127.0.0.1:8100"


@dataclass
class WdaClient:
    """Simple HTTP-based WDA client."""

    base_url: str = DEFAULT_WDA_URL
    timeout: float = 10.0
    session_id: Optional[str] = None

    def _url(self, path: str) -> str:
        """Build a full URL for a WDA path."""
        return f"{self.base_url.rstrip('/')}{path}"

    def _request_json(self, method: str, path: str, payload: Optional[dict[str, Any]] = None) -> dict[str, Any]:
        """Send an HTTP request and parse the WDA JSON response."""
        resp = requests.request(method, self._url(path), json=payload, timeout=self.timeout)
        data = resp.json()
        status = data.get("status")
        if status not in (None, 0, "0"):
            raise WdaError(self._format_error(data, resp.status_code), status_code=resp.status_code)

        return data

    @staticmethod
    def _format_error(data: dict[str, Any], status_code: int) -> str:
        """Format WDA error payloads."""
        message = data.get("value")
        if isinstance(message, dict):
            message = message.get("message") or message.get("error") or message
        return f"WDA error (status={status_code}): {message}"

    def start_session(self, bundle_id: Optional[str] = None) -> str:
        """Start a WDA session (optionally for a bundle id)."""
        caps: dict[str, Any] = {}
        if bundle_id:
            caps["bundleId"] = bundle_id
        payload = {
            "capabilities": {"alwaysMatch": caps},
            "desiredCapabilities": caps,
        }
        data = self._request_json("POST", "/session", payload)
        session_id = data.get("sessionId")
        if not session_id:
            value = data.get("value")
            if isinstance(value, dict):
                session_id = value.get("sessionId")
        if not session_id:
            raise WdaError("WDA did not return a session id")
        self.session_id = session_id
        return session_id

    def find_element(self, using: str, value: str, session_id: Optional[str] = None) -> str:
        """Find an element and return its element id."""
        session_id = session_id or self.session_id
        if not session_id:
            raise WdaError("session_id is required")
        data = self._request_json(
            "POST",
            f"/session/{session_id}/element",
            {"using": using, "value": value},
        )
        element = data.get("value")
        if not isinstance(element, dict):
            raise WdaError("WDA did not return an element")
        element_id = (
            element.get("ELEMENT") or element.get("element-6066-11e4-a52e-4f735466cecf") or element.get("element")
        )
        if not element_id:
            raise WdaError("WDA did not return an element id")
        return element_id

    def click(self, element_id: str, session_id: Optional[str] = None) -> None:
        """Click an element by id."""
        session_id = session_id or self.session_id
        if not session_id:
            raise WdaError("session_id is required")
        self._request_json("POST", f"/session/{session_id}/element/{element_id}/click", {})

    def press_button(self, name: str, session_id: Optional[str] = None) -> None:
        """Press a device button via WDA."""
        normalized = normalize_wda_button_name(name)
        payload = {"name": normalized}
        if session_id:
            try:
                self._request_json("POST", f"/session/{session_id}/wda/pressButton", payload)
            except WdaError as exc:
                if exc.status_code != 404:
                    raise
            else:
                return
            if self._try_keys_endpoint(session_id, normalized):
                return
        if normalized == "home":
            self._request_json("POST", "/wda/homescreen", {})
            return
        raise WdaError("WDA does not support pressButton or keys endpoints", status_code=404)

    def unlock(self, session_id: Optional[str] = None) -> None:
        """Unlock the device via WDA."""
        session_id = session_id or self.session_id
        if session_id:
            try:
                self._request_json("POST", f"/session/{session_id}/wda/unlock", {})
            except WdaError as exc:
                if exc.status_code != 404:
                    raise
            else:
                return
        try:
            self._request_json("POST", "/wda/unlock", {})
        except WdaError as exc:
            if exc.status_code != 404:
                raise
            raise WdaError("WDA does not support unlock endpoint", status_code=404) from exc

    def get_source(self, session_id: Optional[str] = None) -> str:
        """Return the WDA XML source tree."""
        if session_id:
            data = self._request_json("GET", f"/session/{session_id}/source", None)
        else:
            data = self._request_json("GET", "/source", None)
        value = data.get("value")
        if not isinstance(value, str):
            raise WdaError("WDA did not return a source string")
        return value

    def get_screenshot(self, session_id: Optional[str] = None) -> bytes:
        """Return a PNG screenshot as bytes."""
        if session_id:
            data = self._request_json("GET", f"/session/{session_id}/screenshot", None)
        else:
            data = self._request_json("GET", "/screenshot", None)
        value = data.get("value")
        if not isinstance(value, str):
            raise WdaError("WDA did not return a screenshot")
        return base64.b64decode(value)

    def get_status(self) -> dict[str, Any]:
        """Return WDA status payload."""
        return self._request_json("GET", "/status", None)

    def get_window_size(self, session_id: Optional[str] = None) -> dict[str, Any]:
        """Return the current window size."""
        if not session_id:
            raise WdaError("session_id is required")
        data = self._request_json("GET", f"/session/{session_id}/window/size", None)
        value = data.get("value")
        if not isinstance(value, dict):
            raise WdaError("WDA did not return window size")
        return value

    def send_keys(self, text: str, session_id: Optional[str] = None) -> None:
        """Send text input to the focused element."""
        if not session_id:
            raise WdaError("session_id is required")
        payload = {"value": list(text)}
        try:
            self._request_json("POST", f"/session/{session_id}/wda/keys", payload)
        except WdaError as exc:
            if exc.status_code != 404:
                raise
            self._request_json("POST", f"/session/{session_id}/keys", payload)

    def swipe(
        self,
        start_x: int,
        start_y: int,
        end_x: int,
        end_y: int,
        duration: float = 0.2,
        session_id: Optional[str] = None,
    ) -> None:
        """Swipe from one coordinate to another."""
        if not session_id:
            raise WdaError("session_id is required")
        payload = {
            "fromX": start_x,
            "fromY": start_y,
            "toX": end_x,
            "toY": end_y,
            "duration": duration,
        }
        self._request_json("POST", f"/session/{session_id}/wda/dragfromtoforduration", payload)

    def _try_keys_endpoint(self, session_id: Optional[str], normalized: str) -> bool:
        """Try the session keys endpoint for button presses."""
        key = normalize_wda_key_name(normalized)
        payload = {"keys": [key]}
        if session_id:
            try:
                self._request_json("POST", f"/session/{session_id}/wda/keys", payload)
            except WdaError as exc:
                if exc.status_code != 404:
                    raise
            else:
                return True
        return False


@dataclass
class WdaServiceClient:
    """WDA client that uses a LockdownServiceProvider connection."""

    service_provider: LockdownServiceProvider
    port: int = DEFAULT_WDA_PORT
    timeout: float = 10.0
    session_id: Optional[str] = None

    async def _request_json(self, method: str, path: str, payload: Optional[dict[str, Any]] = None) -> dict[str, Any]:
        """Send a WDA request over a service connection and parse JSON."""
        body = b""
        if payload is not None:
            body = json.dumps(payload).encode("utf-8")
        headers = [
            f"{method} {path} HTTP/1.1",
            "Host: localhost",
            "Connection: close",
        ]
        if payload is not None:
            headers.append("Content-Type: application/json")
        headers.append(f"Content-Length: {len(body)}")
        request_bytes = ("\r\n".join(headers) + "\r\n\r\n").encode("utf-8") + body

        if isinstance(self.service_provider, RemoteServiceDiscoveryService):
            conn = await ServiceConnection.create_using_usbmux(
                self.service_provider.udid,
                self.port,
                usbmux_address=getattr(self.service_provider.lockdown, "usbmux_address", None),
            )
        else:
            conn = await self.service_provider.create_service_connection(self.port)
        async with conn:
            await conn.sendall(request_bytes)

            header_bytes, body_prefix = await self._read_until(conn, b"\r\n\r\n")
            header_text = header_bytes.decode("iso-8859-1")
            lines = header_text.split("\r\n")
            status_line = lines[0]
            status_code = int(status_line.split(" ", 2)[1])

            response_headers: dict[str, str] = {}
            for line in lines[1:]:
                if not line:
                    continue
                if ":" not in line:
                    continue
                k, v = line.split(":", 1)
                response_headers[k.strip().lower()] = v.strip()

            content_length = response_headers.get("content-length")
            if content_length is not None:
                try:
                    length = int(content_length)
                except ValueError:
                    length = 0
                if length <= len(body_prefix):
                    body_bytes = body_prefix[:length]
                else:
                    remainder = length - len(body_prefix)
                    body_bytes = body_prefix + await conn.recvall(remainder)
            else:
                chunks = [body_prefix] if body_prefix else []
                while True:
                    chunk = await conn.recv_any(65536)
                    if not chunk:
                        break
                    chunks.append(chunk)
                body_bytes = b"".join(chunks)

        try:
            data = json.loads(body_bytes.decode("utf-8")) if body_bytes else {}
        except ValueError as exc:
            raise WdaError(f"WDA returned non-JSON response (status={status_code})", status_code=status_code) from exc

        if status_code >= 400:
            raise WdaError(WdaClient._format_error(data, status_code), status_code=status_code)

        status = data.get("status")
        if status not in (None, 0, "0"):
            raise WdaError(WdaClient._format_error(data, status_code), status_code=status_code)

        return data

    async def _read_until(self, conn, marker: bytes) -> tuple[bytes, bytes]:
        """Read from a connection until a marker is found."""
        buf = b""
        while marker not in buf:
            chunk = await conn.recv_any(65536)
            if not chunk:
                break
            buf += chunk
        if marker not in buf:
            raise WdaError("WDA response did not contain headers terminator")
        header_bytes, body_prefix = buf.split(marker, 1)
        return header_bytes, body_prefix

    async def start_session(self, bundle_id: Optional[str] = None) -> str:
        """Start a WDA session (optionally for a bundle id)."""
        caps: dict[str, Any] = {}
        if bundle_id:
            caps["bundleId"] = bundle_id
        payload = {
            "capabilities": {"alwaysMatch": caps},
            "desiredCapabilities": caps,
        }
        data = await self._request_json("POST", "/session", payload)
        session_id = data.get("sessionId")
        if not session_id:
            value = data.get("value")
            if isinstance(value, dict):
                session_id = value.get("sessionId")
        if not session_id:
            raise WdaError("WDA did not return a session id")
        self.session_id = session_id
        return session_id

    async def find_element(self, using: str, value: str, session_id: Optional[str] = None) -> str:
        """Find an element and return its element id."""
        session_id = session_id or self.session_id
        if not session_id:
            raise WdaError("session_id is required")
        data = await self._request_json(
            "POST",
            f"/session/{session_id}/element",
            {"using": using, "value": value},
        )
        element = data.get("value")
        if not isinstance(element, dict):
            raise WdaError("WDA did not return an element")
        element_id = (
            element.get("ELEMENT") or element.get("element-6066-11e4-a52e-4f735466cecf") or element.get("element")
        )
        if not element_id:
            raise WdaError("WDA did not return an element id")
        return element_id

    async def click(self, element_id: str, session_id: Optional[str] = None) -> None:
        """Click an element by id."""
        session_id = session_id or self.session_id
        if not session_id:
            raise WdaError("session_id is required")
        await self._request_json("POST", f"/session/{session_id}/element/{element_id}/click", {})

    async def press_button(self, name: str, session_id: Optional[str] = None) -> None:
        """Press a device button via WDA."""
        normalized = normalize_wda_button_name(name)
        payload = {"name": normalized}
        if session_id:
            try:
                await self._request_json("POST", f"/session/{session_id}/wda/pressButton", payload)
            except WdaError as exc:
                if exc.status_code != 404:
                    raise
            else:
                return
            if await self._try_keys_endpoint(session_id, normalized):
                return
        if normalized == "home":
            await self._request_json("POST", "/wda/homescreen", {})
            return
        raise WdaError("WDA does not support pressButton or keys endpoints", status_code=404)

    async def unlock(self, session_id: Optional[str] = None) -> None:
        """Unlock the device via WDA."""
        session_id = session_id or self.session_id
        if session_id:
            try:
                await self._request_json("POST", f"/session/{session_id}/wda/unlock", {})
            except WdaError as exc:
                if exc.status_code != 404:
                    raise
            else:
                return
        try:
            await self._request_json("POST", "/wda/unlock", {})
        except WdaError as exc:
            if exc.status_code != 404:
                raise
            raise WdaError("WDA does not support unlock endpoint", status_code=404) from exc

    async def get_source(self, session_id: Optional[str] = None) -> str:
        """Return the WDA XML source tree."""
        if session_id:
            data = await self._request_json("GET", f"/session/{session_id}/source", None)
        else:
            data = await self._request_json("GET", "/source", None)
        value = data.get("value")
        if not isinstance(value, str):
            raise WdaError("WDA did not return a source string")
        return value

    async def get_screenshot(self, session_id: Optional[str] = None) -> bytes:
        """Return a PNG screenshot as bytes."""
        if session_id:
            data = await self._request_json("GET", f"/session/{session_id}/screenshot", None)
        else:
            data = await self._request_json("GET", "/screenshot", None)
        value = data.get("value")
        if not isinstance(value, str):
            raise WdaError("WDA did not return a screenshot")
        return base64.b64decode(value)

    async def get_status(self) -> dict[str, Any]:
        """Return WDA status payload."""
        return await self._request_json("GET", "/status", None)

    async def get_window_size(self, session_id: Optional[str] = None) -> dict[str, Any]:
        """Return the current window size."""
        if not session_id:
            raise WdaError("session_id is required")
        data = await self._request_json("GET", f"/session/{session_id}/window/size", None)
        value = data.get("value")
        if not isinstance(value, dict):
            raise WdaError("WDA did not return window size")
        return value

    async def send_keys(self, text: str, session_id: Optional[str] = None) -> None:
        """Send text input to the focused element."""
        if not session_id:
            raise WdaError("session_id is required")
        payload = {"value": list(text)}
        try:
            await self._request_json("POST", f"/session/{session_id}/wda/keys", payload)
        except WdaError as exc:
            if exc.status_code != 404:
                raise
            await self._request_json("POST", f"/session/{session_id}/keys", payload)

    async def swipe(
        self,
        start_x: int,
        start_y: int,
        end_x: int,
        end_y: int,
        duration: float = 0.2,
        session_id: Optional[str] = None,
    ) -> None:
        """Swipe from one coordinate to another."""
        if not session_id:
            raise WdaError("session_id is required")
        payload = {
            "fromX": start_x,
            "fromY": start_y,
            "toX": end_x,
            "toY": end_y,
            "duration": duration,
        }
        await self._request_json("POST", f"/session/{session_id}/wda/dragfromtoforduration", payload)

    async def _try_keys_endpoint(self, session_id: Optional[str], normalized: str) -> bool:
        """Try the session keys endpoint for button presses."""
        key = normalize_wda_key_name(normalized)
        payload = {"keys": [key]}
        if session_id:
            try:
                await self._request_json("POST", f"/session/{session_id}/wda/keys", payload)
            except WdaError as exc:
                if exc.status_code != 404:
                    raise
            else:
                return True
        return False


def normalize_wda_button_name(name: str) -> str:
    """Normalize common button aliases to WDA names."""
    key = name.strip().lower().replace("-", "").replace("_", "")
    aliases = {
        "home": "home",
        "volumeup": "volumeUp",
        "volup": "volumeUp",
        "volumeupbutton": "volumeUp",
        "volumedown": "volumeDown",
        "voldown": "volumeDown",
        "volumedownbutton": "volumeDown",
        "lock": "lock",
        "lockscreen": "lock",
        "sleep": "lock",
        "power": "lock",
    }
    return aliases.get(key, name)


def normalize_wda_key_name(name: str) -> str:
    """Normalize common button aliases to WDA key names."""
    key = name.strip().lower().replace("-", "").replace("_", "")
    aliases = {
        "home": "HOME",
        "volumeup": "VOLUME_UP",
        "volup": "VOLUME_UP",
        "volumedown": "VOLUME_DOWN",
        "voldown": "VOLUME_DOWN",
        "lock": "LOCK",
        "lockscreen": "LOCK",
        "sleep": "LOCK",
        "power": "LOCK",
    }
    return aliases.get(key, name)



# extracted from ac2
import logging
import uuid
from typing import Optional

from ipsw_parser.build_identity import BuildIdentity

logger = logging.getLogger(__name__)

SUPPORTED_DATA_TYPES = {
    "BasebandBootData": False,
    "BasebandData": False,
    "BasebandStackData": False,
    "BasebandUpdaterOutputData": False,
    "BootabilityBundle": False,
    "BuildIdentityDict": False,
    "BuildIdentityDictV2": False,
    "DataType": False,
    "DiagData": False,
    "EANData": False,
    "FDRMemoryCommit": False,
    "FDRTrustData": False,
    "FUDData": False,
    "FileData": False,
    "FileDataDone": False,
    "FirmwareUpdaterData": False,
    "GrapeFWData": False,
    "HPMFWData": False,
    "HostSystemTime": True,
    "KernelCache": False,
    "NORData": False,
    "NitrogenFWData": True,
    "OpalFWData": False,
    "OverlayRootDataCount": False,
    "OverlayRootDataForKey": True,
    "PeppyFWData": True,
    "PersonalizedBootObjectV3": False,
    "PersonalizedData": True,
    "ProvisioningData": False,
    "RamdiskFWData": True,
    "RecoveryOSASRImage": True,
    "RecoveryOSAppleLogo": True,
    "RecoveryOSDeviceTree": True,
    "RecoveryOSFileAssetImage": True,
    "RecoveryOSIBEC": True,
    "RecoveryOSIBootFWFilesImages": True,
    "RecoveryOSImage": True,
    "RecoveryOSKernelCache": True,
    "RecoveryOSLocalPolicy": True,
    "RecoveryOSOverlayRootDataCount": False,
    "RecoveryOSRootTicketData": True,
    "RecoveryOSStaticTrustCache": True,
    "RecoveryOSVersionData": True,
    "RootData": False,
    "RootTicket": False,
    "S3EOverride": False,
    "SourceBootObjectV3": False,
    "SourceBootObjectV4": False,
    "SsoServiceTicket": False,
    "StockholmPostflight": False,
    "SystemImageCanonicalMetadata": False,
    "SystemImageData": False,
    "SystemImageRootHash": False,
    "USBCFWData": False,
    "USBCOverride": False,
    "FirmwareUpdaterPreflight": True,
    "ReceiptManifest": True,
    "FirmwareUpdaterDataV2": False,
    "RestoreLocalPolicy": True,
    "AuthInstallCACert": True,
    "OverlayRootDataForKeyIndex": True,
    # Added in iOS 18.0 beta1
    "FirmwareUpdaterDataV3": True,
    "MessageUseStreamedImageFile": True,
    "UpdateVolumeOverlayRootDataCount": True,
    "URLAsset": True,
}

# extracted from ac2
SUPPORTED_MESSAGE_TYPES = {
    "BBUpdateStatusMsg": False,
    "CheckpointMsg": True,
    "CrashLog": True,
    "DataRequestMsg": False,
    "FDRSubmit": True,
    "MsgType": False,
    "PreviousRestoreLogMsg": False,
    "ProgressMsg": False,
    "ProvisioningAck": False,
    "ProvisioningInfo": False,
    "ProvisioningStatusMsg": False,
    "ReceivedFinalStatusMsg": False,
    "RestoredCrash": True,
    "StatusMsg": False,
    # Added in iOS 18.0 beta1
    "AsyncDataRequestMsg": True,
    "AsyncWait": True,
    "RestoreAttestation": True,
}


class RestoreOptions:
    def __init__(
        self,
        firmware_preflight_info=None,
        sep=None,
        macos_variant=None,
        build_identity: BuildIdentity = None,
        restore_boot_args=None,
        spp=None,
        restore_behavior: Optional[str] = None,
        msp=None,
    ):
        self.AutoBootDelay = 0

        try:
            if firmware_preflight_info is not None:
                bbus = dict(firmware_preflight_info)
                bbus.pop("FusingStatus")
                bbus.pop("PkHash")
                self.BBUpdaterState = bbus

                nonce = firmware_preflight_info.get("Nonce")
                if nonce is not None:
                    self.BasebandNonce = nonce
        except KeyError as e:
            logger.warning(f"Skipping addition of firmware_preflight_info due to: {e}")

        self.SupportedDataTypes = SUPPORTED_DATA_TYPES
        self.SupportedMessageTypes = SUPPORTED_MESSAGE_TYPES

        # FIXME: Should be adjusted for update behaviors
        if macos_variant:
            self.AddSystemPartitionPadding = True
            self.AllowUntetheredRestore = False
            self.AuthInstallEnableSso = False

            macos_variant = build_identity.macos_variant
            if macos_variant is not None:
                self.AuthInstallRecoveryOSVariant = macos_variant

            self.AuthInstallRestoreBehavior = restore_behavior
            self.AutoBootDelay = 0
            self.BasebandUpdaterOutputPath = True
            self.DisableUserAuthentication = True
            self.FitSystemPartitionToContent = True
            self.FlashNOR = True
            self.FormatForAPFS = True
            self.FormatForLwVM = False
            self.InstallDiags = False
            self.InstallRecoveryOS = True
            self.MacOSSwapPerformed = True
            self.MacOSVariantPresent = True
            self.MinimumBatteryVoltage = 0  # FIXME: Should be adjusted for M1 macbooks (if needed)
            self.RecoveryOSUnpack = True
            self.ShouldRestoreSystemImage = True
            self.SkipPreflightPersonalization = False
            self.UpdateBaseband = True

            # FIXME: I don't know where this number comes from yet.
            #  It seems like it matches this part of the build identity:
            # 	<key>OSVarContentSize</key>
            # 	<integer>573751296</integer>
            # It did work with multiple macOS versions
            self.recoveryOSPartitionSize = 58201
            if msp:
                self.SystemPartitionSize = msp
        else:
            self.BootImageType = "UserOrInternal"
            self.DFUFileType = "RELEASE"
            self.DataImage = False
            self.FirmwareDirectory = "."
            self.FlashNOR = True
            self.KernelCacheType = "Release"
            self.NORImageType = "production"
            self.RestoreBundlePath = "/tmp/Per2.tmp"
            self.SystemImageType = "User"
            self.UpdateBaseband = False

            # Added for iOS 18.0 beta1
            self.HostHasFixFor99053849 = True
            self.SystemImageFormat = "AEAWrappedDiskImage"
            self.WaitForDeviceConnectionToFinishStateMachine = False
            self.SupportedAsyncDataTypes = {
                "BasebandData": False,
                "RecoveryOSASRImage": False,
                "StreamedImageDecryptionKey": False,
                "SystemImageData": False,
                "URLAsset": True,
            }

            if sep is not None:
                required_capacity = sep.get("RequiredCapacity")
                if required_capacity:
                    logger.debug(f"TZ0RequiredCapacity: {required_capacity}")
                    self.TZ0RequiredCapacity = required_capacity

            self.PersonalizedDuringPreflight = True

        self.RootToInstall = False
        self.UUID = str(uuid.uuid4()).upper()
        self.CreateFilesystemPartitions = True
        self.SystemImage = True

        if restore_boot_args is not None:
            self.RestoreBootArgs = restore_boot_args

        if spp:
            spp = dict(spp)
        else:
            spp = {
                "1024": 1280,
                "128": 1280,
                "16": 160,
                "256": 1280,
                "32": 320,
                "512": 1280,
                "64": 640,
                "768": 1280,
                "8": 80,
            }
        self.SystemPartitionPadding = spp

    def to_dict(self):
        return self.__dict__



import uuid
from typing import Any, Optional

from pymobiledevice3.exceptions import CoreDeviceError
from pymobiledevice3.remote.remote_service import RemoteService
from pymobiledevice3.remote.xpc_message import XpcInt64Type, XpcUInt64Type


def _generate_core_device_version_dict(version: str) -> dict:
    version_components = version.split(".")
    return {
        "components": [XpcUInt64Type(component) for component in version_components],
        "originalComponentsCount": XpcInt64Type(len(version_components)),
        "stringValue": version,
    }


CORE_DEVICE_VERSION = _generate_core_device_version_dict("325.3")


class CoreDeviceService(RemoteService):
    async def invoke(self, feature_identifier: str, input_: Optional[dict] = None) -> Any:
        if input_ is None:
            input_ = {}
        response = await self.service.send_receive_request({
            "CoreDevice.CoreDeviceDDIProtocolVersion": XpcInt64Type(0),
            "CoreDevice.action": {},
            "CoreDevice.coreDeviceVersion": CORE_DEVICE_VERSION,
            "CoreDevice.deviceIdentifier": str(uuid.uuid4()),
            "CoreDevice.featureIdentifier": feature_identifier,
            "CoreDevice.input": input_,
            "CoreDevice.invocationIdentifier": str(uuid.uuid4()),
        })
        output = response.get("CoreDevice.output")
        if output is None:
            raise CoreDeviceError(f"Failed to invoke: {feature_identifier}. Got error: {response}")
        return output



import asyncio
import dataclasses
import logging
import typing
from typing import Any, Optional

from pymobiledevice3.dtx import ConnectionAwareQueue, DTXService, PInt32, dtx_method, dtx_on_invoke
from pymobiledevice3.dtx_service import DtxService
from pymobiledevice3.exceptions import DisableMemoryLimitError
from pymobiledevice3.osu.os_utils import get_os_utils

OSUTIL = get_os_utils()


@dataclasses.dataclass
class OutputReceivedEvent:
    pid: int
    date: int
    message: str

    @classmethod
    def create(cls, message: list[Any]) -> "OutputReceivedEvent":
        def _value(v: Any) -> Any:
            return getattr(v, "value", v)

        msg_value = _value(message[0])
        pid_value = _value(message[1])
        timestamp_value = _value(message[2])
        try:
            date = OSUTIL.parse_timestamp(timestamp_value)
        except (ValueError, OSError):
            date = None

        return cls(pid=pid_value, date=date, message=msg_value)


class ProcessControlService(DTXService):
    IDENTIFIER = "com.apple.instruments.server.services.processcontrol"

    def __init__(self, ctx):
        super().__init__(ctx)
        self.output_events: asyncio.Queue[list[Any]] = ConnectionAwareQueue()

    @dtx_method("sendSignal:toPid:")
    async def send_signal_to_pid_(self, sig: int, pid: int) -> Any: ...

    @dtx_method("requestDisableMemoryLimitsForPid:")
    async def request_disable_memory_limits_for_pid_(self, pid: PInt32) -> bool: ...

    @dtx_method("killPid:", expects_reply=False)
    async def kill_pid_(self, pid: int) -> None: ...

    @dtx_method("processIdentifierForBundleIdentifier:")
    async def process_identifier_for_bundle_identifier_(self, app_bundle_identifier: str) -> int: ...

    @dtx_method("launchSuspendedProcessWithDevicePath:bundleIdentifier:environment:arguments:options:")
    async def launch_suspended_process_with_device_path_bundle_identifier_environment_arguments_options_(
        self, device_path: str, bundle_id: str, environment: dict, arguments: list, options: dict
    ) -> int: ...

    @dtx_on_invoke("outputReceived:fromProcess:atTime:")
    async def _on_output_received(self, message: str, pid: int, timestamp: Any) -> None:
        await self.output_events.put([message, pid, timestamp])

    @dtx_method
    async def startObservingPid_(self, pid: int) -> None: ...

    @dtx_method
    async def stopObservingPid_(self, pid: int) -> None: ...

    @dtx_on_invoke("processWithPID:terminatedWithExitCode:orCrashingSignal:")
    async def _on_process_terminated(self, pid: int, exit_code: Optional[int], crashing_signal: Optional[int]) -> None:
        logger = self._ctx.get("logger") or logging.getLogger(__name__)
        logger.warning(
            f"Process with PID {pid} terminated with exit code {exit_code} or crashing signal {crashing_signal}"
        )


class ProcessControl(DtxService[ProcessControlService]):
    async def connect(self):
        await super().connect()
        self._provider.dtx.ctx["logger"] = self.logger.getChild("dtx")

    async def signal(self, pid: int, sig: int):
        """
        Send signal to process
        :param pid: PID of process to send signal.
        :param sig: SIGNAL to send
        """
        return await self.service.send_signal_to_pid_(sig, pid)

    async def disable_memory_limit_for_pid(self, pid: int) -> None:
        """
        Waive memory limit for a given pid
        :param pid: process id.
        """
        if not await self.service.request_disable_memory_limits_for_pid_(pid):
            raise DisableMemoryLimitError()

    async def kill(self, pid: int):
        """
        Kill a process.
        :param pid: PID of process to kill.
        """
        await self.service.kill_pid_(pid)

    async def process_identifier_for_bundle_identifier(self, app_bundle_identifier: str) -> int:
        return await self.service.process_identifier_for_bundle_identifier_(app_bundle_identifier)

    async def launch(
        self,
        bundle_id: str,
        arguments=None,
        kill_existing: bool = True,
        start_suspended: bool = False,
        environment: Optional[dict] = None,
        extra_options: Optional[dict] = None,
    ) -> int:
        """
        Launch a process.
        :param bundle_id: Bundle id of the process.
        :param list arguments: List of argument to pass to process.
        :param kill_existing: Whether to kill an existing instance of this process.
        :param start_suspended: Same as WaitForDebugger.
        :param environment: Environment variables to pass to process.
        :param extra_options: Extra options to pass to process.
        :return: PID of created process.
        """
        arguments = [] if arguments is None else arguments
        environment = {} if environment is None else environment
        options = {
            "StartSuspendedKey": start_suspended,
            "KillExisting": kill_existing,
        }
        if extra_options:
            options.update(extra_options)
        result = await self.service.launch_suspended_process_with_device_path_bundle_identifier_environment_arguments_options_(
            "", bundle_id, environment, arguments, options
        )
        assert result
        return result

    async def __aiter__(self) -> typing.AsyncGenerator[OutputReceivedEvent, None]:
        while True:
            value = await self.service.output_events.get()
            yield OutputReceivedEvent.create(value)



import asyncio
import contextlib
import hashlib
import logging
from typing import Optional

from ipsw_parser.ipsw import IPSW
from usb import USBError

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.irecv import IRecv, Mode
from pymobiledevice3.lockdown import create_using_usbmux
from pymobiledevice3.restore.base_restore import BaseRestore, Behavior
from pymobiledevice3.restore.consts import lpol_file
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.tss import TSSRequest, TSSResponse

RESTORE_VARIANT_ERASE_INSTALL = "Erase Install (IPSW)"
RESTORE_VARIANT_UPGRADE_INSTALL = "Upgrade Install (IPSW)"
RESTORE_VARIANT_MACOS_RECOVERY_OS = "macOS Customer"


class Recovery(BaseRestore):
    def __init__(self, ipsw: IPSW, device: Device, tss: Optional[dict] = None, behavior: Behavior = Behavior.Update):
        super().__init__(ipsw, device, tss, behavior)
        self.tss_localpolicy = None
        self.tss_recoveryos_root_ticket = None
        self.restore_boot_args = None

    async def reconnect_irecv(self, is_recovery=None):
        self.logger.debug("waiting for device to reconnect...")
        self.device.set_irecv(IRecv(ecid=await self.device.get_ecid(), is_recovery=is_recovery))
        self.logger.debug(f"connected mode: {self.device.irecv.mode}")

    def get_preboard_manifest(self):
        overrides = {
            "@APTicket": True,
            "ApProductionMode": 0,
            "ApSecurityDomain": 0,
        }

        parameters = {
            "ApProductionMode": False,
            "ApSecurityMode": False,
            "ApSupportsImg4": True,
        }

        self.populate_tss_request_from_manifest(parameters)

        tss = TSSRequest()
        tss.add_common_tags(parameters, overrides)

        parameters["_OnlyFWComponents"] = True

        tss.add_ap_tags(parameters)

        return tss.img4_create_local_manifest(build_identity=self.build_identity)

    async def get_tss_response(self) -> TSSResponse:
        # populate parameters
        parameters = {}

        parameters["ApECID"] = await self.device.get_ecid()
        ap_nonce = await self.device.get_ap_nonce()
        if ap_nonce is not None:
            parameters["ApNonce"] = ap_nonce

        sep_nonce = await self.device.get_sep_nonce()
        if sep_nonce is not None:
            parameters["ApSepNonce"] = sep_nonce

        parameters["ApProductionMode"] = True

        if await self.device.get_is_image4_supported():
            parameters["ApSecurityMode"] = True
            parameters["ApSupportsImg4"] = True
        else:
            parameters["ApSupportsImg4"] = False

        self.populate_tss_request_from_manifest(parameters)

        tss = TSSRequest()
        tss.add_common_tags(parameters)
        tss.add_ap_tags(parameters)

        # TODO: This break updating iPhone 15P Pro. Consider re-adding it once we figure out what went wrong
        # build_manifest_info = self.build_identity['Info']
        # for manifest_property in build_manifest_info.get('RequestManifestProperties', []):
        #     tss.add_tags({manifest_property: build_manifest_info[manifest_property]})

        # add personalized parameters
        if await self.device.get_is_image4_supported():
            tss.add_ap_img4_tags(parameters)
        else:
            tss.add_ap_img3_tags(parameters)

        # normal mode; request baseband ticket as well
        if self.device.lockdown is not None:
            pinfo = await self.device.get_firmware_preflight_info()
            if pinfo:
                self.logger.debug("adding firmware preflight info")

                node = pinfo.get("Nonce")
                if node is not None:
                    parameters["BbNonce"] = node

                node = pinfo.get("ChipID")
                if node is not None:
                    parameters["BbChipID"] = node

                node = pinfo.get("CertID")
                if node is not None:
                    parameters["BbGoldCertId"] = node

                node = pinfo.get("ChipSerialNo")
                if node is not None:
                    parameters["BbSNUM"] = node

                    # add baseband parameters
                    tss.add_baseband_tags(parameters)

                    euiccchipid = pinfo.get("EUICCChipID")
                    if euiccchipid:
                        self.logger.debug("adding EUICCChipID info")
                        parameters["eUICC,ChipID"] = euiccchipid

                        if euiccchipid >= 5:
                            node = pinfo.get("EUICCCSN")
                            if node is not None:
                                parameters["eUICC,EID"] = node

                            node = pinfo.get("EUICCCertIdentifier")
                            if node is not None:
                                parameters["eUICC,RootKeyIdentifier"] = node

                            node = pinfo.get("EUICCGoldNonce")
                            if node is not None:
                                parameters["EUICCGoldNonce"] = node

                            node = pinfo.get("EUICCMainNonce")
                            if node is not None:
                                parameters["EUICCMainNonce"] = node

                            tss.add_vinyl_tags(parameters)

        # send request and grab response
        return await tss.send_receive()

    async def get_local_policy_tss_response(self):
        # populate parameters
        parameters = {
            "ApECID": await self.device.get_ecid(),
            "Ap,LocalBoot": False,
            "ApProductionMode": True,
        }

        ap_nonce = await self.device.get_ap_nonce()
        if ap_nonce:
            parameters["ApNonce"] = ap_nonce

        sep_nonce = await self.device.get_sep_nonce()

        if sep_nonce:
            parameters["ApSepNonce"] = sep_nonce

        if await self.device.get_is_image4_supported():
            parameters["ApSecurityMode"] = True
            parameters["ApSupportsImg4"] = True
        else:
            parameters["ApSupportsImg4"] = False

        self.populate_tss_request_from_manifest(parameters)

        # Add Ap,LocalPolicy
        lpol = {
            "Digest": hashlib.sha384(lpol_file).digest(),
            "Trusted": True,
        }

        parameters["Ap,LocalPolicy"] = lpol

        # Add Ap,NextStageIM4MHash
        # Get previous TSS ticket
        ticket = self.tss.ap_img4_ticket
        # Hash it and add it as Ap,NextStageIM4MHash
        parameters["Ap,NextStageIM4MHash"] = hashlib.sha384(ticket).digest()

        # create basic request
        request = TSSRequest()

        # add common tags from manifest
        request.add_local_policy_tags(parameters)

        return await request.send_receive()

    async def get_recoveryos_root_ticket_tss_response(self):
        # populate parameters
        parameters = {
            "ApECID": await self.device.get_ecid(),
            "Ap,LocalBoot": False,
            "ApProductionMode": True,
        }

        ap_nonce = await self.device.get_ap_nonce()
        if ap_nonce:
            parameters["ApNonce"] = ap_nonce

        sep_nonce = await self.device.get_sep_nonce()

        if sep_nonce:
            parameters["ApSepNonce"] = sep_nonce

        if await self.device.get_is_image4_supported():
            parameters["ApSecurityMode"] = True
            parameters["ApSupportsImg4"] = True
        else:
            parameters["ApSupportsImg4"] = False

        self.populate_tss_request_from_manifest(parameters)

        # create basic request
        # Adds @HostPlatformInfo, @VersionInfo, @UUID
        request = TSSRequest()

        # add common tags from manifest
        # Adds Ap,OSLongVersion, AppNonce, @ApImg4Ticket
        request.add_ap_img4_tags(parameters)

        # add AP tags from manifest
        request.add_common_tags(parameters)

        # add AP tags from manifest
        # Fills digests & co
        request.add_ap_recovery_tags(parameters)

        return await request.send_receive()

    async def fetch_tss_record(self) -> TSSResponse:
        await self.ensure_image4_supported()

        if self.ipsw.build_manifest.build_major > 8 and await self.device.get_ap_nonce() is None:
            # the first nonce request with older firmware releases can fail, and it's OK
            self.logger.info("NOTE: Unable to get nonce from device")

        self.tss = await self.get_tss_response()

        if self.macos_variant:
            self.tss_localpolicy = await self.get_local_policy_tss_response()
            self.tss_recoveryos_root_ticket = await self.get_recoveryos_root_ticket_tss_response()
        else:
            recovery_variant = self.build_identity["Info"].get("RecoveryVariant")
            if recovery_variant is not None:
                self.tss_recoveryos_root_ticket = await self.get_tss_response()

        return self.tss

    async def send_component(self, name: str):
        # Use a specific TSS ticket for the Ap,LocalPolicy component
        data = None
        tss = self.tss
        if name == "Ap,LocalPolicy":
            tss = self.tss_localpolicy
            # If Ap,LocalPolicy => Inject an empty policy
            data = lpol_file

        data = await self.get_personalized_data(name, data=data, tss=tss)
        self.logger.info(f"Sending {name} ({len(data)} bytes)...")
        self.device.irecv.send_buffer(data)

    async def send_component_and_command(self, name, command):
        await self.send_component(name)
        self.device.irecv.send_command(command)

    async def send_ibec(self):
        component = "iBEC"
        await self.send_component(component)
        self.device.irecv.send_command("go", b_request=1)
        self.device.irecv.ctrl_transfer(0x21, 1)

    async def send_applelogo(self, allow_missing=True):
        component = "RestoreLogo"

        if not self.build_identity.has_component(component):
            if allow_missing:
                logging.warning(f"build_identity has no {component}")
                return
            else:
                raise PyMobileDevice3Exception(f"missing component: {component}")

        await self.send_component(component)
        self.device.irecv.send_command("setpicture 4")
        self.device.irecv.send_command("bgcolor 0 0 0")

    async def send_loaded_by_iboot(self):
        manifest = self.build_identity["Manifest"]
        for key, node in manifest.items():
            iboot = node["Info"].get("IsLoadedByiBoot", False)
            iboot_stg1 = node["Info"].get("IsLoadedByiBootStage1", False)

            assert isinstance(iboot, bool)
            assert isinstance(iboot_stg1, bool)

            if iboot and not iboot_stg1:
                self.logger.debug(f"{key} is loaded by iBoot")
                await self.send_component_and_command(key, "firmware")

    async def send_iboot_stage1_components(self):
        manifest = self.build_identity["Manifest"]
        for key, node in manifest.items():
            iboot = node["Info"].get("IsLoadedByiBoot", False)
            iboot_stg1 = node["Info"].get("IsLoadedByiBootStage1", False)

            assert isinstance(iboot, bool)
            assert isinstance(iboot_stg1, bool)

            if iboot and iboot_stg1:
                self.logger.debug(f"{key} is loaded by iBoot Stage 1")
                await self.send_component_and_command(key, "firmware")

    async def send_ramdisk(self):
        component = "RestoreRamDisk"
        ramdisk_size = self.device.irecv.getenv("ramdisk-size")
        self.logger.info(f"ramdisk-size: {ramdisk_size}")

        await self.send_component(component)
        ramdisk_delay = self.device.irecv.getenv("ramdisk-delay")
        self.logger.info(f"ramdisk-delay: {ramdisk_delay}")

        self.device.irecv.send_command("ramdisk")

        await asyncio.sleep(2)

    async def send_kernelcache(self):
        component = "RestoreKernelCache"

        await self.send_component(component)
        with contextlib.suppress(USBError):
            self.device.irecv.ctrl_transfer(0x21, 1)

        if self.restore_boot_args:
            self.device.irecv.send_command(f"setenv boot-args {self.restore_boot_args}")

        with contextlib.suppress(USBError):
            self.device.irecv.send_command("bootx", b_request=1)

    def set_autoboot(self, enable: bool):
        self.device.irecv.set_autoboot(enable)

    async def enter_restore(self):
        if self.macos_variant:
            self.restore_boot_args = "rd=md0 nand-enable-reformat=1 -progress -restore"
        elif self.ipsw.build_manifest.build_major >= 8:
            self.restore_boot_args = "rd=md0 nand-enable-reformat=1 -progress"

        # upload data to make device boot restore mode

        # Recovery Mode Environment:
        build_version = None
        while not build_version:
            self.logger.debug("build-version not yet supported. reconnecting...")
            await asyncio.sleep(1)

            # sometimes we manage to connect before iBEC actually started running
            build_version = self.device.irecv.getenv("build-version")
            await self.reconnect_irecv()

        self.logger.info(f"iBoot build-version={build_version}")

        build_style = self.device.irecv.getenv("build-style")
        self.logger.info(f"iBoot build-style={build_style}")

        radio_error = self.device.irecv.getenv("radio-error")
        if radio_error:
            radio_error = int(radio_error)
            self.logger.info(f"radio-error: {radio_error}")
            radio_error_string = self.device.irecv.getenv("radio-error-string")
            if radio_error_string:
                self.logger.info(f"radio-error-string: {radio_error_string}")

        self.set_autoboot(False)

        # send logo and show it
        await self.send_applelogo()

        # send components loaded by iBoot
        await self.send_loaded_by_iboot()

        # send ramdisk and run it
        await self.send_ramdisk()

        # send devicetree and load it
        await self.send_component_and_command("RestoreDeviceTree", "devicetree")

        if self.build_identity.has_component("RestoreSEP"):
            # attempt to send rsepfirmware and load it, otherwise continue
            with contextlib.suppress(USBError):
                await self.send_component_and_command("RestoreSEP", "rsepfirmware")

        await self.send_kernelcache()

    async def dfu_enter_recovery(self) -> None:
        await self.send_component("iBSS")
        await self.reconnect_irecv()

        if "SRTG" in self.device.irecv._device_info:
            raise PyMobileDevice3Exception("Device failed to enter recovery")

        if self.build_identity.build_manifest.build_major > 8:
            old_nonce = self.device.irecv.ap_nonce

            # reconnect
            await self.reconnect_irecv()
            nonce = self.device.irecv.ap_nonce

            if old_nonce != nonce:
                # Welcome iOS5. We have to re-request the TSS with our nonce.
                self.tss = await self.get_tss_response()

            self.device.irecv.set_configuration(1)

            # Now, before sending iBEC, we must send necessary firmwares on new versions.
            if self.macos_variant:
                # Without this empty policy file & its special signature, iBEC won't start.
                await self.send_component_and_command("Ap,LocalPolicy", "lpolrestore")
                await self.send_iboot_stage1_components()
                self.device.irecv.set_autoboot(False)
                self.device.irecv.send_command("setenvnp boot-args rd=md0 nand-enable-reformat=1 -progress -restore")
                await self.send_applelogo(allow_missing=False)

            mode = self.device.irecv.mode
            # send iBEC
            await self.send_component("iBEC")

            if self.device.irecv and mode.is_recovery:
                await asyncio.sleep(1)
                self.device.irecv.send_command("go", b_request=1)

                if self.build_identity.build_manifest.build_major < 20:
                    with contextlib.suppress(USBError):
                        self.device.irecv.ctrl_transfer(0x21, 1, timeout=5000)

                self.logger.debug("Waiting for device to disconnect...")
                await asyncio.sleep(10)

        self.logger.debug("Waiting for device to reconnect in recovery mode...")
        await self.reconnect_irecv(is_recovery=True)

    async def boot_ramdisk(self) -> None:
        if self.tss is None:
            self.logger.info("fetching TSS record")
            await self.fetch_tss_record()

        if self.device.lockdown:
            # normal mode
            self.logger.info("going into Recovery")

            # In case lockdown has disconnected while waiting for a ticket
            self.device.set_lockdown(await create_using_usbmux(serial=self.device.lockdown.udid, connection_type="USB"))
            await self.device.lockdown.enter_recovery()

            self.device.set_lockdown(None)
            self.device.set_irecv(IRecv(await self.device.get_ecid()))
            await self.reconnect_irecv()

        if self.device.irecv.mode == Mode.DFU_MODE:
            # device is currently in DFU mode, place it into recovery mode
            await self.dfu_enter_recovery()

        elif self.device.irecv.mode.is_recovery:
            # now we load the iBEC
            with contextlib.suppress(USBError):
                await self.send_ibec()

            await self.reconnect_irecv(is_recovery=True)

        self.logger.info("device booted into recovery")

        # now finally do the magic to put the device into restore mode
        await self.enter_restore()



import ast
import datetime
import logging
import plistlib
from pathlib import Path
from typing import Annotated, Literal, Optional

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import (
    NoAutoPairServiceProviderDep,
    ServiceProviderDep,
    async_command,
    print_json,
    sudo_required,
)
from pymobiledevice3.cli.remote import tunnel_task
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.common import TunnelProtocol
from pymobiledevice3.remote.tunnel_service import CoreDeviceTunnelProxy
from pymobiledevice3.services.heartbeat import HeartbeatService
from pymobiledevice3.utils import run_in_loop

logger = logging.getLogger(__name__)

cli = InjectingTyper(
    name="lockdown",
    help="Pair/Unpair device or access other lockdown services",
    no_args_is_help=True,
)


@cli.command("recovery")
@async_command
async def lockdown_recovery(service_provider: ServiceProviderDep) -> None:
    """enter recovery"""
    print_json(await service_provider.enter_recovery())


@cli.command("service")
async def lockdown_service(service_provider: ServiceProviderDep, service_name: str) -> None:
    """send-receive raw service messages with a given service name"""
    service = run_in_loop(service_provider.start_lockdown_service(service_name))
    service.shell()


@cli.command("developer-service")
async def lockdown_developer_service(service_provider: ServiceProviderDep, service_name: str) -> None:
    """send-receive raw service messages with a given developer service name"""
    service = run_in_loop(service_provider.start_lockdown_developer_service(service_name))
    service.shell()


@cli.command("info")
def lockdown_info(service_provider: ServiceProviderDep) -> None:
    """query all lockdown values"""
    print_json(service_provider.all_values)


@cli.command("get")
@async_command
async def lockdown_get(
    service_provider: ServiceProviderDep, domain: Optional[str] = None, key: Optional[str] = None
) -> None:
    """query lockdown values by their domain and key names"""
    print_json(await service_provider.get_value(domain=domain, key=key))


@cli.command("set")
@async_command
async def lockdown_set(
    service_provider: ServiceProviderDep,
    value: str,
    domain: Optional[str] = None,
    key: Optional[str] = None,
) -> None:
    """set a lockdown value using python's ast.literal_eval()"""
    print_json(await service_provider.set_value(value=ast.literal_eval(value), domain=domain, key=key))


@cli.command("remove")
@async_command
async def lockdown_remove(service_provider: ServiceProviderDep, domain: str, key: str) -> None:
    """remove a domain/key pair"""
    print_json(await service_provider.remove_value(domain=domain, key=key))


@cli.command("unpair")
@async_command
async def lockdown_unpair(service_provider: NoAutoPairServiceProviderDep, host_id: Optional[str] = None) -> None:
    """unpair from connected device"""
    await service_provider.unpair(host_id=host_id)


@cli.command("pair")
@async_command
async def lockdown_pair(service_provider: NoAutoPairServiceProviderDep) -> None:
    """pair device"""
    await service_provider.pair()


@cli.command("pair-supervised")
@async_command
async def lockdown_pair_supervised(
    service_provider: NoAutoPairServiceProviderDep,
    keybag: Annotated[
        Path,
        typer.Argument(file_okay=True, dir_okay=False, exists=True),
    ],
) -> None:
    """pair supervised device"""
    await service_provider.pair_supervised(keybag)


@cli.command("save-pair-record")
def lockdown_save_pair_record(service_provider: NoAutoPairServiceProviderDep, output: Path) -> None:
    """save pair record to specified location"""
    if service_provider.pair_record is None:
        logger.error("no pairing record was found")
        return
    output.write_bytes(plistlib.dumps(service_provider.pair_record))


@cli.command("date")
@async_command
async def lockdown_date(service_provider: ServiceProviderDep) -> None:
    """get device date"""
    timestamp = await service_provider.get_value(key="TimeIntervalSince1970")
    print(datetime.datetime.fromtimestamp(timestamp))


@cli.command("heartbeat")
@async_command
async def lockdown_heartbeat(service_provider: ServiceProviderDep) -> None:
    """start heartbeat service"""
    await HeartbeatService(service_provider).start()


@cli.command("language")
@async_command
async def lockdown_language(
    service_provider: ServiceProviderDep, language: Annotated[Optional[str], typer.Argument()] = None
) -> None:
    """Get/Set current language settings"""
    if language is not None:
        await service_provider.set_language(language)
    print_json(await service_provider.get_language())


@cli.command("locale")
@async_command
async def lockdown_locale(
    service_provider: ServiceProviderDep, locale: Annotated[Optional[str], typer.Argument()] = None
) -> None:
    """Get/Set current language settings"""
    if locale is not None:
        await service_provider.set_locale(locale)
    print_json(await service_provider.get_locale())


@cli.command("device-name")
@async_command
async def lockdown_device_name(service_provider: ServiceProviderDep, new_name: Optional[str] = None) -> None:
    """get/set current device name"""
    if new_name:
        await service_provider.set_value(new_name, key="DeviceName")
    else:
        print(f"{await service_provider.get_value(key='DeviceName')}")


@cli.command("wifi-connections")
@async_command
async def lockdown_wifi_connections(
    service_provider: ServiceProviderDep, state: Optional[Literal["on", "off"]] = None
) -> None:
    """get/set wifi connections state"""
    if not state:
        # show current state
        print_json({"EnableWifiConnections": await service_provider.get_enable_wifi_connections()})
    else:
        # enable/disable
        await service_provider.set_enable_wifi_connections(state == "on")


async def async_cli_start_tunnel(service_provider: LockdownServiceProvider, script_mode: bool) -> None:
    await tunnel_task(
        await CoreDeviceTunnelProxy.create(service_provider),
        script_mode=script_mode,
        secrets=None,
        protocol=TunnelProtocol.TCP,
    )


@cli.command("start-tunnel")
@sudo_required
@async_command
async def cli_start_tunnel(
    service_provider: ServiceProviderDep,
    script_mode: Annotated[
        bool,
        typer.Option(help="Show only HOST and port number to allow easy parsing from external shell scripts"),
    ] = False,
) -> None:
    """start tunnel"""
    await async_cli_start_tunnel(service_provider, script_mode)


@cli.command("assistive-touch")
@async_command
async def lockdown_assistive_touch(
    service_provider: ServiceProviderDep, state: Optional[Literal["on", "off"]] = None
) -> None:
    """get/set assistive touch icon state (visibility)"""
    if not state:
        print_json({"AssistiveTouchEnabledByiTunes": await service_provider.get_assistive_touch()})
    else:
        # enable/disable
        await service_provider.set_assistive_touch(state == "on")



import logging
from typing import Callable, Optional

logger = logging.getLogger(__name__)

try:
    from pymobiledevice3.remote.tunnel_service import RemotePairingQuicTunnel, start_tunnel

    MAX_IDLE_TIMEOUT = RemotePairingQuicTunnel.MAX_IDLE_TIMEOUT
except ImportError:
    start_tunnel: Optional[Callable] = None
    MAX_IDLE_TIMEOUT = None

GENERAL_IMPORT_ERROR = """Failed to import `start_tunnel`.
Please file an issue at:
https://github.com/doronz88/pymobiledevice3/issues/new?assignees=&labels=&projects=&template=bug_report.md&title=

Also, please supply with a traceback of the following python line:

from pymobiledevice3.remote.tunnel_service import start_tunnel
"""


def verify_tunnel_imports() -> bool:
    if start_tunnel is not None:
        return True
    logger.error(GENERAL_IMPORT_ERROR)
    return False



import asyncio
import contextlib
import json
import logging
import uuid
from dataclasses import dataclass, fields
from enum import Enum
from typing import Optional, Union

from pymobiledevice3.exceptions import (
    LaunchingApplicationError,
    RemoteAutomationNotEnabledError,
    WebInspectorNotEnabledError,
)
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.service_connection import ServiceConnection
from pymobiledevice3.services.web_protocol.automation_session import AutomationSession
from pymobiledevice3.services.web_protocol.inspector_session import InspectorSession
from pymobiledevice3.services.web_protocol.session_protocol import SessionProtocol

SAFARI = "com.apple.mobilesafari"


def key_to_pid(key: str) -> int:
    return int(key.split(":")[1])


class WirTypes(Enum):
    AUTOMATION = "WIRTypeAutomation"
    ITML = "WIRTypeITML"
    JAVASCRIPT = "WIRTypeJavaScript"
    PAGE = "WIRTypePage"
    SERVICE_WORKER = "WIRTypeServiceWorker"
    WEB = "WIRTypeWeb"
    WEB_PAGE = "WIRTypeWebPage"
    AUTOMATICALLY_PAUSE = "WIRAutomaticallyPause"


class AutomationAvailability(Enum):
    NOT_AVAILABLE = "WIRAutomationAvailabilityNotAvailable"
    AVAILABLE = "WIRAutomationAvailabilityAvailable"
    UNKNOWN = "WIRAutomationAvailabilityUnknown"


@dataclass
class Page:
    id_: int
    type_: WirTypes
    web_url: str = ""
    web_title: str = ""
    automation_is_paired_key: bool = False
    automation_name: str = ""
    automation_version: str = ""
    automation_session_id: str = ""
    automation_connection_id: str = ""

    @classmethod
    def from_page_dictionary(cls, page_dict: dict) -> "Page":
        p = cls(page_dict["WIRPageIdentifierKey"], WirTypes(page_dict["WIRTypeKey"]))
        if p.type_ in (WirTypes.WEB, WirTypes.WEB_PAGE):
            p.web_title = page_dict["WIRTitleKey"]
            p.web_url = page_dict["WIRURLKey"]
        if p.type_ == WirTypes.AUTOMATION:
            p.automation_is_paired_key = page_dict["WIRAutomationTargetIsPairedKey"]
            p.automation_name = page_dict["WIRAutomationTargetNameKey"]
            p.automation_version = page_dict["WIRAutomationTargetVersionKey"]
            p.automation_session_id = page_dict["WIRSessionIdentifierKey"]
            if "WIRConnectionIdentifierKey" in page_dict:
                p.automation_connection_id = page_dict["WIRConnectionIdentifierKey"]
        return p

    def update(self, page_dict: dict):
        new_p = self.from_page_dictionary(page_dict)
        for field in fields(self):
            setattr(self, field.name, getattr(new_p, field.name))

    def __str__(self):
        return f"id: {self.id_}, title: {self.web_title}, url: {self.web_url}"


@dataclass
class Application:
    id_: str
    bundle: str
    pid: int
    name: str
    availability: AutomationAvailability
    active: int
    proxy: bool
    ready: bool
    host: str = ""

    @classmethod
    def from_application_dictionary(cls, app_dict) -> "Application":
        return cls(
            app_dict["WIRApplicationIdentifierKey"],
            app_dict["WIRApplicationBundleIdentifierKey"],
            key_to_pid(app_dict["WIRApplicationIdentifierKey"]),
            app_dict["WIRApplicationNameKey"],
            AutomationAvailability(app_dict["WIRAutomationAvailabilityKey"]),
            app_dict["WIRIsApplicationActiveKey"],
            app_dict["WIRIsApplicationProxyKey"],
            app_dict["WIRIsApplicationReadyKey"],
            app_dict.get("WIRHostApplicationIdentifierKey", ""),
        )


@dataclass
class ApplicationPage:
    application: Application
    page: Page

    def __str__(self) -> str:
        return f"<{self.application.name}({self.application.pid}) TYPE:{self.page.type_.value} URL:{self.page.web_url}>"


class WebinspectorService:
    SERVICE_NAME = "com.apple.webinspector"
    RSD_SERVICE_NAME = "com.apple.webinspector.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            self.service_name = self.SERVICE_NAME
        else:
            self.service_name = self.RSD_SERVICE_NAME

        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service: Optional[ServiceConnection] = None
        self.connection_id = str(uuid.uuid4()).upper()
        self.state = None
        self.connected_application = {}
        self.application_pages = {}
        self.wir_message_results = {}
        self.wir_events = []
        self.receive_handlers = {
            "_rpc_reportCurrentState:": self._handle_report_current_state,
            "_rpc_reportConnectedApplicationList:": self._handle_report_connected_application_list,
            "_rpc_reportConnectedDriverList:": self._handle_report_connected_driver_list,
            "_rpc_applicationSentListing:": self._handle_application_sent_listing,
            "_rpc_applicationUpdated:": self._handle_application_updated,
            "_rpc_applicationConnected:": self._handle_application_connected,
            "_rpc_applicationSentData:": self._handle_application_sent_data,
            "_rpc_applicationDisconnected:": self._handle_application_disconnected,
        }
        self._recv_task: Optional[asyncio.Task] = None

    async def connect(self, timeout: Optional[Union[float, int]] = None):
        self.service = await self.lockdown.start_lockdown_service(self.service_name)
        await self._report_identifier()
        try:
            await self._handle_recv(await asyncio.wait_for(self._recv_message(), timeout))
        except asyncio.TimeoutError as e:
            raise WebInspectorNotEnabledError from e
        self._recv_task = asyncio.create_task(self._receiving_task())

    async def close(self):
        if self._recv_task is not None:
            self._recv_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._recv_task

        if self.service is not None:
            await self.service.close()

    async def _recv_message(self):
        while True:
            try:
                return await self.service.recv_plist()
            except asyncio.IncompleteReadError:
                await asyncio.sleep(0)

    async def _receiving_task(self):
        while True:
            await self._handle_recv(await self._recv_message())

    async def automation_session(self, app: Application) -> AutomationSession:
        if self.state == "WIRAutomationAvailabilityNotAvailable":
            raise RemoteAutomationNotEnabledError()
        session_id = str(uuid.uuid4()).upper()
        await self._forward_automation_session_request(session_id, app.id_)
        await self._forward_get_listing(app.id_)
        page = await self._wait_for_page(session_id)
        await self._forward_socket_setup(session_id, app.id_, page.id_)
        await self._forward_get_listing(app.id_)
        while not page.automation_connection_id:
            await asyncio.sleep(0)
        return AutomationSession(SessionProtocol(self, session_id, app, page))

    async def inspector_session(self, app: Application, page: Page) -> InspectorSession:
        session_id = str(uuid.uuid4()).upper()
        return await InspectorSession.create(
            SessionProtocol(self, session_id, app, page, method_prefix=""),
            wait_target=page.type_ != WirTypes.JAVASCRIPT,
        )

    async def get_open_pages(self) -> dict:
        apps = {}
        await asyncio.gather(*[self._forward_get_listing(app) for app in self.connected_application])
        for app in self.connected_application:
            if self.application_pages.get(app, False):
                apps[self.connected_application[app].name] = self.application_pages[app].values()
        return apps

    async def get_open_application_pages(self, timeout: float) -> list[ApplicationPage]:
        # Query all connected applications
        await self._get_connected_applications()

        # Give some time for `webinspectord` to reply with all inspectable applications
        await asyncio.sleep(timeout)

        result = []
        for app in self.connected_application:
            if self.application_pages.get(app, False):
                for page in self.application_pages[app].values():
                    result.append(ApplicationPage(self.connected_application[app], page))
        return result

    async def open_app(self, bundle: str, timeout: Union[float, int] = 3) -> Application:
        await self._request_application_launch(bundle)
        await self.get_open_pages()
        try:
            return await asyncio.wait_for(self._wait_for_application(bundle), timeout)
        except TimeoutError as e:
            raise LaunchingApplicationError() from e

    async def send_socket_data(self, session_id: str, app_id: str, page_id: int, data: dict):
        await self._forward_socket_data(session_id, app_id, page_id, data)

    async def setup_inspector_socket(self, session_id: str, app_id: str, page_id: int):
        await self._forward_socket_setup(session_id, app_id, page_id, pause=False)

    def find_page_id(self, page_id: str) -> tuple[Application, Page]:
        for app_id in self.application_pages:
            for page in self.application_pages[app_id]:
                if page == page_id:
                    return self.connected_application[app_id], self.application_pages[app_id][page_id]
        raise KeyError(f"Page with id {page_id} not found")

    async def flush_input(self, duration: Union[float, int] = 0):
        return await asyncio.sleep(duration)

    async def _handle_recv(self, plist):
        await self.receive_handlers[plist["__selector"]](plist["__argument"])

    async def _handle_report_current_state(self, arg):
        self.state = arg["WIRAutomationAvailabilityKey"]

    async def _handle_report_connected_application_list(self, arg):
        self.connected_application = {}
        for key, application in arg["WIRApplicationDictionaryKey"].items():
            self.connected_application[key] = Application.from_application_dictionary(application)

            # Immediately also query the application pages
            await self._forward_get_listing(self.connected_application[key].id_)

    async def _handle_report_connected_driver_list(self, arg):
        pass

    async def _handle_application_sent_listing(self, arg):
        if arg["WIRApplicationIdentifierKey"] in self.application_pages:
            # Update existing application pages
            for id_, page in arg["WIRListingKey"].items():
                if id_ in self.application_pages[arg["WIRApplicationIdentifierKey"]]:
                    self.application_pages[arg["WIRApplicationIdentifierKey"]][id_].update(page)
                else:
                    self.application_pages[arg["WIRApplicationIdentifierKey"]][id_] = Page.from_page_dictionary(page)
        else:
            # Add new application pages
            pages = {}
            for id_, page in arg["WIRListingKey"].items():
                pages[id_] = Page.from_page_dictionary(page)
            self.application_pages[arg["WIRApplicationIdentifierKey"]] = pages

    async def _handle_application_updated(self, arg):
        app = Application.from_application_dictionary(arg)
        self.connected_application[app.id_] = app

    async def _handle_application_connected(self, arg):
        app = Application.from_application_dictionary(arg)
        self.connected_application[app.id_] = app

    async def _handle_application_sent_data(self, arg):
        response = json.loads(arg["WIRMessageDataKey"])

        if "id" in response:
            self.wir_message_results[response["id"]] = response
        else:
            self.wir_events.append(response)

    async def _handle_application_disconnected(self, arg):
        self.connected_application.pop(arg["WIRApplicationIdentifierKey"], None)
        self.application_pages.pop(arg["WIRApplicationIdentifierKey"], None)

    async def _report_identifier(self):
        await self._send_message("_rpc_reportIdentifier:")

    async def _forward_get_listing(self, app_id):
        self.logger.debug(f"Listing app with id {app_id}")
        await self._send_message("_rpc_forwardGetListing:", {"WIRApplicationIdentifierKey": app_id})

    async def _request_application_launch(self, bundle: str):
        await self._send_message("_rpc_requestApplicationLaunch:", {"WIRApplicationBundleIdentifierKey": bundle})

    async def _get_connected_applications(self) -> None:
        await self._send_message("_rpc_getConnectedApplications:", {})

    async def _forward_automation_session_request(self, session_id: str, app_id: str):
        await self._send_message(
            "_rpc_forwardAutomationSessionRequest:",
            {
                "WIRApplicationIdentifierKey": app_id,
                "WIRSessionCapabilitiesKey": {
                    "org.webkit.webdriver.webrtc.allow-insecure-media-capture": True,
                    "org.webkit.webdriver.webrtc.suppress-ice-candidate-filtering": False,
                },
                "WIRSessionIdentifierKey": session_id,
            },
        )

    async def _forward_socket_setup(self, session_id: str, app_id: str, page_id: int, pause: bool = True):
        message = {
            "WIRApplicationIdentifierKey": app_id,
            "WIRPageIdentifierKey": page_id,
            "WIRSenderKey": session_id,
            "WIRMessageDataTypeChunkSupportedKey": 0,
        }
        if not pause:
            message["WIRAutomaticallyPause"] = False
        await self._send_message("_rpc_forwardSocketSetup:", message)

    async def _forward_socket_data(self, session_id: str, app_id: str, page_id: int, data: dict):
        await self._send_message(
            "_rpc_forwardSocketData:",
            {
                "WIRApplicationIdentifierKey": app_id,
                "WIRPageIdentifierKey": page_id,
                "WIRSessionIdentifierKey": session_id,
                "WIRSenderKey": session_id,
                "WIRSocketDataKey": json.dumps(data).encode(),
            },
        )

    async def _forward_indicate_web_view(self, app_id: str, page_id: int, enable: bool):
        (
            await self._send_message("_rpc_forwardIndicateWebView"),
            {
                "WIRApplicationIdentifierKey": app_id,
                "WIRPageIdentifierKey": page_id,
                "WIRIndicateEnabledKey": enable,
            },
        )

    async def _send_message(self, selector: str, args=None):
        if args is None:
            args = {}
        args["WIRConnectionIdentifierKey"] = self.connection_id
        await self.service.send_plist({"__selector": selector, "__argument": args})

    def _page_by_automation_session(self, session_id: str) -> Page:
        for app_id in self.application_pages:
            for page in self.application_pages[app_id]:
                if page.type_ == WirTypes.AUTOMATION and page.automation_session_id == session_id:
                    return page
        raise KeyError(f"Automation session with id {session_id} not found")

    async def _wait_for_page(self, session_id: str):
        while True:
            for app in self.application_pages.values():
                for page in app.values():
                    if page.type_ == WirTypes.AUTOMATION and page.automation_session_id == session_id:
                        return page
            await asyncio.sleep(0)

    async def _wait_for_application(self, bundle: str = "", app_id: str = "") -> Application:
        while True:
            for app in self.connected_application.values():
                if bundle and app.bundle == bundle:
                    return app
                if app_id and app.id_ == app_id:
                    return app
            await asyncio.sleep(0)



import json
from io import StringIO

import pytest
import typer

from pymobiledevice3.cli.developer.dvt.sysmon import process as process_module
from pymobiledevice3.cli.developer.dvt.sysmon.process import (
    ProcessSelectionMode,
    _describe_process,
    _describe_processes,
    _duration_elapsed,
    _format_byte_count,
    _get_process_identifier,
    _humanize_process_values,
    _matches_filters,
    _matches_selected_process,
    _parse_process_filters,
    _process_sort_key,
    _select_process_from_candidates,
    _select_process_from_snapshot,
    _select_process_from_sysmon,
    _select_process_output_keys,
    _serialize_process,
    _should_skip_first_snapshot,
    _validate_process_keys,
    _write_json,
    _write_process,
    iter_processes,
    sysmon_process_monitor_threshold_task,
)


class _FakeSysmontap:
    def __init__(self, snapshots):
        self._snapshots = snapshots

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return None

    async def iter_processes(self):
        for snapshot in self._snapshots:
            yield snapshot


class _FakeDvtProvider:
    def __init__(self, service_provider):
        self.service_provider = service_provider

    async def __aenter__(self):
        return object()

    async def __aexit__(self, exc_type, exc, tb):
        return None


def test_parse_process_filters_groups_values_by_key():
    assert _parse_process_filters(["name=abc", "name=def", "pid=7"]) == {
        "name": ["abc", "def"],
        "pid": ["7"],
    }


@pytest.mark.parametrize("raw_filter", ["no_value", "=no_key"])
def test_parse_process_filters_rejects_invalid_input(raw_filter):
    with pytest.raises(typer.BadParameter):
        _parse_process_filters([raw_filter])


def test_validate_process_keys_rejects_unknown_keys():
    with pytest.raises(typer.BadParameter, match="does not have the following keys"):
        _validate_process_keys({"pid": 1, "name": "abc"}, ["pid", "missing"])


def test_matches_filters_requires_all_keys_and_any_value_per_key():
    process = {"pid": 123, "name": "abc", "realAppName": "Calculator"}
    assert _matches_filters(process, {"pid": ["123"], "name": ["abc", "def"]})
    assert not _matches_filters(process, {"pid": ["999"]})
    assert not _matches_filters(process, {"bundleIdentifier": ["com.example.app"]})


def test_select_process_output_keys_returns_process_when_no_keys_requested():
    process = {"pid": 1, "name": "abc"}
    assert _select_process_output_keys(process, None) == process


def test_select_process_output_keys_filters_selected_keys():
    assert _select_process_output_keys({"pid": 1, "name": "abc", "cpuUsage": 7.5}, ["pid", "name"]) == {
        "pid": 1,
        "name": "abc",
    }


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (10, "10B"),
        (1024, "1.0KB"),
        (10 * 1024, "10KB"),
        (1024 * 1024, "1.0MB"),
    ],
)
def test_format_byte_count(value, expected):
    assert _format_byte_count(value) == expected


def test_humanize_process_values_formats_only_known_byte_fields():
    assert _humanize_process_values({"physFootprint": 2048, "name": "abc", "cpuUsage": 2.0}) == {
        "physFootprint": "2.0KB",
        "name": "abc",
        "cpuUsage": 2.0,
    }


def test_serialize_process_filters_selected_keys():
    serialized = _serialize_process({"pid": 1, "name": "abc", "cpuUsage": 7.5}, ["pid", "name"])
    assert serialized["pid"] == 1
    assert serialized["name"] == "abc"
    assert "cpuUsage" not in serialized
    assert "timestamp" in serialized


def test_serialize_process_humanizes_selected_values():
    serialized = _serialize_process({"physFootprint": 2048, "cpuUsage": 1.5}, ["physFootprint"], human=True)
    assert serialized["physFootprint"] == "2.0KB"
    assert "cpuUsage" not in serialized
    assert "timestamp" in serialized


def test_write_process_writes_jsonl_to_output_stream():
    out = StringIO()
    _write_process(out, {"pid": 1})
    assert out.getvalue() == '{"pid": 1}\n'


def test_write_process_prints_json_when_out_is_none(monkeypatch):
    captured = {}

    def fake_print_json(value):
        captured["value"] = value

    monkeypatch.setattr(process_module, "print_json", fake_print_json)

    _write_process(None, {"pid": 1})

    assert captured["value"] == {"pid": 1}


def test_write_json_writes_formatted_json_to_output_stream():
    out = StringIO()
    _write_json(out, [{"pid": 1}])
    assert json.loads(out.getvalue()) == [{"pid": 1}]


def test_write_json_prints_json_when_out_is_none(monkeypatch):
    captured = {}

    def fake_print_json(value):
        captured["value"] = value

    monkeypatch.setattr(process_module, "print_json", fake_print_json)

    _write_json(None, [{"pid": 1}])

    assert captured["value"] == [{"pid": 1}]


def test_describe_process_includes_pid_ppid_name():
    assert _describe_process({"pid": 7, "ppid": 1, "name": "abc", "comm": "abc"}) == "pid=7, ppid=1, name=abc"


def test_describe_process_falls_back_to_comm():
    assert _describe_process({"pid": 7, "ppid": 1, "comm": "abc"}) == "pid=7, ppid=1, name=abc"


def test_describe_process_falls_back_to_unknown():
    assert _describe_process({"pid": 7, "ppid": 1}) == "pid=7, ppid=1, name=<unknown>"


def test_describe_processes_joins_descriptions():
    assert (
        _describe_processes([
            {"pid": 7, "ppid": 1, "name": "abc"},
            {"pid": 8, "ppid": 2, "name": "def"},
        ])
        == "pid=7, ppid=1, name=abc; pid=8, ppid=2, name=def"
    )


def test_duration_elapsed_false_when_duration_is_none():
    assert _duration_elapsed(10.0, None) is False


@pytest.mark.parametrize(
    ("start_time", "duration_ms", "current_time", "expected"),
    [
        (1.0, 1000, 1.5, False),
        (1.0, 1000, 2.0, True),
        (1.0, 0, 1.0, True),
    ],
)
def test_duration_elapsed(monkeypatch, start_time, duration_ms, current_time, expected):
    fake_loop = type("FakeLoop", (), {"time": lambda self: current_time})()
    monkeypatch.setattr(process_module.asyncio, "get_running_loop", lambda: fake_loop)
    assert _duration_elapsed(start_time, duration_ms) is expected


@pytest.mark.parametrize(
    ("keys", "expected"),
    [
        (None, True),
        (["pid"], False),
        (["pid", "cpuUsage"], True),
    ],
)
def test_should_skip_first_snapshot(keys, expected):
    assert _should_skip_first_snapshot(keys) is expected


@pytest.mark.asyncio
async def test_iter_processes_skips_first_snapshot_when_requested():
    snapshots = [
        [{"pid": 10, "ppid": 1, "name": "first-snapshot"}],
        [{"pid": 20, "ppid": 2, "name": "second-snapshot"}],
        [{"pid": 30, "ppid": 3, "name": "third-snapshot"}],
    ]
    sysmon = _FakeSysmontap(snapshots)

    iterated_snapshots = [snapshot async for snapshot in iter_processes(sysmon, skip_first_snapshot=True)]

    assert iterated_snapshots == snapshots[1:]


@pytest.mark.asyncio
async def test_iter_processes_empty_when_only_warmup_snapshot_exists():
    sysmon = _FakeSysmontap([[{"pid": 10, "ppid": 1, "name": "first-snapshot"}]])

    snapshots = [snapshot async for snapshot in iter_processes(sysmon, skip_first_snapshot=True)]

    assert snapshots == []


@pytest.mark.asyncio
async def test_iter_processes_keeps_first_snapshot_when_cpu_usage_not_required():
    snapshots = [
        [{"pid": 10, "ppid": 1, "name": "first-snapshot"}],
        [{"pid": 20, "ppid": 2, "name": "second-snapshot"}],
    ]

    sysmon = _FakeSysmontap(snapshots)

    iterated_snapshots = [snapshot async for snapshot in iter_processes(sysmon, skip_first_snapshot=False)]

    assert iterated_snapshots == snapshots


@pytest.mark.asyncio
async def test_sysmon_process_monitor_threshold_task_skips_first_snapshot(monkeypatch):
    async def fake_create(_dvt):
        return _FakeSysmontap([
            [{"pid": 10, "cpuUsage": 99.0}],
            [{"pid": 20, "cpuUsage": 1.5}],
        ])

    monkeypatch.setattr(process_module, "DvtProvider", _FakeDvtProvider)
    monkeypatch.setattr(process_module.Sysmontap, "create", fake_create)

    out = StringIO()

    await sysmon_process_monitor_threshold_task(object(), threshold=0.0, keys=["pid"], out=out, duration=1)

    lines = [json.loads(line) for line in out.getvalue().splitlines() if line.strip()]
    assert len(lines) == 1
    assert lines[0]["pid"] == 20


def test_process_sort_key_normalizes_missing_values():
    assert _process_sort_key({"name": "abc"}) == (-1, -1, "abc")


def test_select_process_from_candidates_returns_single_match():
    process = {"pid": 7, "name": "abc"}
    assert _select_process_from_candidates([process], ProcessSelectionMode.PROMPT) == process


def test_select_process_from_candidates_first_is_deterministic():
    first = {"pid": 3, "name": "abc", "startAbsTime": 10}
    second = {"pid": 2, "name": "abc", "startAbsTime": 20}
    assert _select_process_from_candidates([second, first], ProcessSelectionMode.FIRST) == first


def test_select_process_from_candidates_last_is_deterministic():
    first = {"pid": 3, "name": "abc", "startAbsTime": 10}
    last = {"pid": 2, "name": "abc", "startAbsTime": 20}
    assert _select_process_from_candidates([last, first], ProcessSelectionMode.LAST) == last


def test_select_process_from_candidates_multiple_non_tty_raises(monkeypatch):
    monkeypatch.setattr(process_module.sys, "stdin", type("FakeStdin", (), {"isatty": lambda self: False})())

    with pytest.raises(typer.BadParameter, match='Re-run with "--choose first", "--choose last"'):
        _select_process_from_candidates(
            [{"pid": 1, "ppid": 0, "name": "a"}, {"pid": 2, "ppid": 0, "name": "b"}],
            ProcessSelectionMode.PROMPT,
        )


def test_select_process_from_candidates_prompt_uses_sorted_choices(monkeypatch):
    monkeypatch.setattr(process_module.sys, "stdin", type("FakeStdin", (), {"isatty": lambda self: True})())

    captured = {}

    def fake_prompt_selection(choices, message, idx=False):
        captured["choices"] = choices
        captured["message"] = message
        captured["idx"] = idx
        return 1

    monkeypatch.setattr(process_module, "prompt_selection", fake_prompt_selection)

    processes = [
        {"pid": 20, "ppid": 2, "name": "later", "startAbsTime": 20},
        {"pid": 10, "ppid": 1, "name": "earlier", "startAbsTime": 10},
    ]

    selected = _select_process_from_candidates(processes, ProcessSelectionMode.PROMPT)

    assert selected == processes[0]
    assert captured["choices"] == [
        "pid=10, ppid=1, name=earlier",
        "pid=20, ppid=2, name=later",
    ]
    assert captured["message"] == "Choose process to monitor"
    assert captured["idx"] is True


def test_get_process_identifier_prefers_unique_id_then_start_time_then_pid():
    assert _get_process_identifier({"uniqueID": 11, "startAbsTime": 22, "pid": 33}) == ("uniqueID", 11)
    assert _get_process_identifier({"startAbsTime": 22, "pid": 33}) == ("startAbsTime", 22)
    assert _get_process_identifier({"pid": 33}) == ("pid", 33)


def test_matches_selected_process_uses_identifier_tuple():
    assert _matches_selected_process({"uniqueID": 11}, ("uniqueID", 11))
    assert not _matches_selected_process({"uniqueID": 12}, ("uniqueID", 11))


def test_select_process_from_snapshot_filters_current_snapshot_only():
    process_snapshot = [
        {"pid": 10, "ppid": 1, "name": "alpha"},
        {"pid": 20, "ppid": 1, "name": "beta"},
    ]

    selected = _select_process_from_snapshot(process_snapshot, {"name": ["beta"]}, ProcessSelectionMode.FIRST)

    assert selected == {"pid": 20, "ppid": 1, "name": "beta"}


def test_select_process_from_snapshot_rejects_unknown_filter_keys():
    with pytest.raises(typer.BadParameter, match="does not have the following keys"):
        _select_process_from_snapshot(
            [{"pid": 10, "ppid": 1, "name": "alpha"}],
            {"missing": ["value"]},
            ProcessSelectionMode.FIRST,
        )


def test_select_process_from_snapshot_raises_when_no_match():
    with pytest.raises(typer.BadParameter, match="current snapshot"):
        _select_process_from_snapshot(
            [{"pid": 10, "ppid": 1, "name": "alpha"}],
            {"name": ["beta"]},
            ProcessSelectionMode.FIRST,
        )


@pytest.mark.asyncio
async def test_select_process_from_sysmon_skips_first_snapshot(monkeypatch):
    async def fake_create(_dvt):
        return _FakeSysmontap([
            [{"pid": 10, "ppid": 1, "name": "first-snapshot"}],
            [{"pid": 20, "ppid": 2, "name": "second-snapshot"}],
        ])

    monkeypatch.setattr(process_module.Sysmontap, "create", fake_create)

    selected = await _select_process_from_sysmon(
        object(), {"name": ["second-snapshot"]}, None, ProcessSelectionMode.FIRST
    )

    assert selected == {"pid": 20, "ppid": 2, "name": "second-snapshot"}


@pytest.mark.asyncio
async def test_select_process_from_sysmon_doesnt_skip_first_snapshot_when_cpu_usage_not_requested(monkeypatch):
    async def fake_create(_dvt):
        return _FakeSysmontap([
            [{"pid": 10, "ppid": 1, "name": "first-snapshot"}],
            [{"pid": 20, "ppid": 2, "name": "second-snapshot"}],
        ])

    monkeypatch.setattr(process_module.Sysmontap, "create", fake_create)

    selected = await _select_process_from_sysmon(
        object(), {"name": ["first-snapshot"]}, ["pid", "name"], ProcessSelectionMode.FIRST
    )

    assert selected == {"pid": 10, "ppid": 1, "name": "first-snapshot"}


@pytest.mark.asyncio
async def test_select_process_from_sysmon_raises_when_no_usable_snapshot(monkeypatch):
    async def fake_create(_dvt):
        return _FakeSysmontap([])

    monkeypatch.setattr(process_module.Sysmontap, "create", fake_create)

    with pytest.raises(typer.BadParameter, match="Failed to collect a process snapshot"):
        await _select_process_from_sysmon(object(), {}, None, ProcessSelectionMode.FIRST)



from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService


class ScreenshotService(LockdownService):
    SERVICE_NAME = "com.apple.mobile.screenshotr"

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        super().__init__(lockdown, self.SERVICE_NAME, is_developer_service=True)
        self._did_handshake = False

    async def _handshake(self) -> None:
        if self._did_handshake:
            return
        dl_message_version_exchange = await self.service.recv_plist()
        version_major = dl_message_version_exchange[1]
        await self.service.send_plist(["DLMessageVersionExchange", "DLVersionsOk", version_major])
        dl_message_device_ready = await self.service.recv_plist()
        if dl_message_device_ready[0] != "DLMessageDeviceReady":
            raise PyMobileDevice3Exception("Screenshotr didn't return ready state")
        self._did_handshake = True

    async def take_screenshot(self) -> bytes:
        await self._handshake()
        await self.service.send_plist(["DLMessageProcessMessage", {"MessageType": "ScreenShotRequest"}])
        response = await self.service.recv_plist()

        assert len(response) == 2
        assert response[0] == "DLMessageProcessMessage"

        if response[1].get("MessageType") == "ScreenShotReply":
            return response[1]["ScreenShotData"]

        raise PyMobileDevice3Exception(f"invalid response: {response}")



import binascii
import contextlib
import logging
import math
import struct
import time
from enum import Enum
from typing import Optional

from tqdm import trange
from usb.core import Device, USBError, find
from usb.util import get_string

from pymobiledevice3.exceptions import IRecvError, IRecvNoDeviceConnectedError, PyMobileDevice3Exception
from pymobiledevice3.irecv_devices import IRECV_DEVICES, IRecvDevice

USB_TIMEOUT = 10000
IRECV_TRANSFER_SIZE_RECOVERY = 0x8000
IRECV_TRANSFER_SIZE_DFU = 0x800


class Mode(Enum):
    RECOVERY_MODE_1 = 0x1280
    RECOVERY_MODE_2 = 0x1281
    RECOVERY_MODE_3 = 0x1282
    RECOVERY_MODE_4 = 0x1283
    WTF_MODE = 0x1222
    DFU_MODE = 0x1227

    @classmethod
    def has_value(cls, value):
        return any(value == m.value for m in cls)

    @classmethod
    def get_mode_from_value(cls, value):
        """
        :rtype: Mode
        """
        for m in cls:
            if value == m.value:
                return m
        return None

    @property
    def is_recovery(self):
        return self not in (self.WTF_MODE, self.DFU_MODE)


CPFM_FLAG_SECURITY_MODE = 1 << 0
CPFM_FLAG_PRODUCTION_MODE = 1 << 1

IBOOT_FLAG_IMAGE4_AWARE = 1 << 2
IBOOT_FLAG_EFFECTIVE_SECURITY_MODE = 1 << 3
IBOOT_FLAG_EFFECTIVE_PRODUCTION_MODE = 1 << 4

APPLE_VENDOR_ID = 0x05AC

logger = logging.getLogger(__name__)


class IRecv:
    def __init__(self, ecid=None, timeout=0xFFFFFFFF, is_recovery=None):
        self.mode: Optional[Mode] = None
        self._device_info = {}
        self._device: Optional[Device] = None
        self._reinit(ecid=ecid, timeout=timeout, is_recovery=is_recovery)

    @property
    def ecid(self):
        return int(self._device_info["ECID"], 16)

    @property
    def ibfl(self):
        return int(self._device_info["IBFL"], 16)

    @property
    def chip_id(self):
        return int(self._device_info["CPID"], 16)

    @property
    def board_id(self):
        return int(self._device_info["BDID"], 16)

    @property
    def serial_number(self) -> int:
        return self._device_info["SRNM"]

    @property
    def iboot_version(self) -> str:
        return self._device_info["SRTG"]

    @property
    def is_image4_supported(self):
        return self.ibfl & IBOOT_FLAG_IMAGE4_AWARE

    @property
    def _irecv_device(self) -> IRecvDevice:
        for device in IRECV_DEVICES:
            if device.board_id == self.board_id and device.chip_id == self.chip_id:
                return device
        raise KeyError(f"failed to find device of: board_id: {self.board_id} chip_id: {self.chip_id}")

    @property
    def product_type(self):
        return self._irecv_device.product_type

    @property
    def hardware_model(self):
        return self._irecv_device.hardware_model

    @property
    def display_name(self):
        return self._irecv_device.display_name

    @property
    def status(self):
        return self.ctrl_transfer(0xA1, 3, data_or_wLength=b"\x00" * 6)[4]

    def set_interface_altsetting(self, interface=None, alternate_setting=None):
        logger.debug(f"set_interface_altsetting: {interface} {alternate_setting}")
        if interface == 1:
            self._device.set_interface_altsetting(interface=interface, alternate_setting=alternate_setting)

    def set_configuration(self, configuration=None):
        logger.debug(f"set_configuration: {configuration}")
        try:
            if self._device.get_active_configuration().bConfigurationValue == configuration:
                return
        except USBError:
            pass

        self._device.set_configuration(configuration=configuration)

    def ctrl_transfer(self, bmRequestType, bRequest, timeout=USB_TIMEOUT, **kwargs):
        return self._device.ctrl_transfer(bmRequestType, bRequest, timeout=timeout, **kwargs)

    def send_buffer(self, buf: bytes):
        packet_size = IRECV_TRANSFER_SIZE_RECOVERY if self.mode.is_recovery else IRECV_TRANSFER_SIZE_DFU

        # initiate transfer
        if self.mode.is_recovery:
            self.ctrl_transfer(0x41, 0)
        else:
            response = self.ctrl_transfer(0xA1, 5, data_or_wLength=1)
            state = response[0]
            logger.debug(f"irecv state: {state}")
            if state == 2:
                # DFU IDLE
                pass
            elif state == 10:
                self.ctrl_transfer(0x21, 4)
                raise PyMobileDevice3Exception("DFU ERROR, issuing CLRSTATUS")
            else:
                self.ctrl_transfer(0x21, 6)
                raise PyMobileDevice3Exception(f"Unexpected state {state}, issuing ABORT")

        crc = -1

        num_packets = math.ceil(len(buf) / packet_size)

        for _offset in trange(0, len(buf), packet_size, dynamic_ncols=True):
            # Use bulk transfer for recovery mode and control transfer for DFU and WTF mode
            chunk = buf[_offset : _offset + packet_size]
            packet_index = _offset // packet_size

            if self.mode.is_recovery:
                n = self._device.write(0x04, chunk, timeout=USB_TIMEOUT)
                if n != len(chunk):
                    raise OSError("failed to upload data")
            else:
                if _offset + packet_size >= len(buf):
                    # last packet

                    # calculate crc of all sent data
                    crc = binascii.crc32(buf, crc)

                    # add crc of dfu_xbuf (salted value)
                    dfu_xbuf = bytearray([0xFF, 0xFF, 0xFF, 0xFF, 0xAC, 0x05, 0x00, 0x01, 0x55, 0x46, 0x44, 0x10])
                    crc = binascii.crc32(dfu_xbuf, crc)

                    crc_chunk = dfu_xbuf + struct.pack("<I", crc)

                    if len(chunk) + 16 > packet_size:
                        # crc exceeds the max allowed packet size
                        self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=chunk)
                        self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=crc_chunk)
                    else:
                        self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=chunk + crc_chunk)
                else:
                    self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=chunk)

        if not self.mode.is_recovery:
            logger.debug("waiting for status == 5")
            while self.status != 5:
                time.sleep(1)

            self.ctrl_transfer(0x21, 1, wValue=num_packets, wIndex=0)

            for _offset in range(2):
                # i know it's not used but idevicerestore does that also
                _ = self.status

            self.reset()

    def reset(self):
        try:
            logger.debug("resetting usb device")
            logger.info("If the restore hangs here, disconnect & reconnect the device")
            self._device.reset()
        except USBError:
            pass

        self._reinit(ecid=self.ecid)

    def send_command(self, cmd: str, timeout=USB_TIMEOUT, b_request=0):
        self._device.ctrl_transfer(0x40, b_request, 0, 0, cmd.encode() + b"\0", timeout=timeout)

    def getenv(self, name):
        try:
            self.send_command(f"getenv {name}")
        except USBError:
            return None
        return bytearray(self._device.ctrl_transfer(0xC0, 0, 0, 0, 255))

    def set_autoboot(self, enable: bool):
        self.send_command(f"setenv auto-boot {str(enable).lower()}")
        self.send_command("saveenv")

    def reboot(self):
        with contextlib.suppress(USBError):
            self.send_command("reboot")

    def _reinit(self, ecid=None, timeout=0xFFFFFFFF, is_recovery=None):
        self._device = None
        self._device_info = {}
        self.mode = None
        self._find(ecid=ecid, timeout=timeout, is_recovery=is_recovery)
        if self._device is None:
            raise IRecvNoDeviceConnectedError("Failed to find a connected iDevice")
        self._populate_device_info()

        self.ap_nonce = self._copy_nonce_with_tag("NONC")
        self.sep_nonce = self._copy_nonce_with_tag("SNON")

        self.set_configuration(1)

        if self.mode.is_recovery:
            self.set_interface_altsetting(0, 0)
            if self.mode.value > Mode.RECOVERY_MODE_2.value:
                self.set_interface_altsetting(1, 1)
        else:
            self.set_interface_altsetting(0, 0)

    def _copy_nonce_with_tag(self, tag):
        if tag in get_string(self._device, 1):
            return binascii.unhexlify(get_string(self._device, 1).split(f"{tag}:")[1].split(" ")[0])
        else:
            return None

    def _find(self, ecid=None, timeout=0xFFFFFFFF, is_recovery=None):
        start = time.time()
        end = start + timeout
        while (self._device is None) and (time.time() < end):
            for device in find(find_all=True):
                try:
                    if device.manufacturer is None:
                        continue
                    if not device.manufacturer.startswith("Apple"):
                        continue

                    mode = Mode.get_mode_from_value(device.idProduct)
                    if mode is None:
                        # not one of Apple's special modes
                        continue

                    if is_recovery is not None and mode.is_recovery != is_recovery:
                        continue

                    if self._device is not None:
                        raise IRecvError("More then one connected device was found connected in recovery mode")
                    self._device = device
                    self.mode = mode
                    self._populate_device_info()

                    if ecid is not None:
                        found_ecid = int(self._device_info["ECID"], 16)
                        if found_ecid != ecid:
                            # wrong device - move on
                            self._device = None
                            continue
                except ValueError:
                    continue

    def _populate_device_info(self):
        for component in self._device.serial_number.split(" "):
            k, v = component.split(":")
            if k in ("SRNM", "SRTG") and "[" in v:
                # trim the `[]`
                v = v[1:-1]
            self._device_info[k] = v

    def __str__(self):
        return str(self._device_info)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self._device


def main():
    print(IRecv())


if __name__ == "__main__":
    main()



from pymobiledevice3.services.web_protocol.alert import Alert
from pymobiledevice3.services.web_protocol.automation_session import By
from pymobiledevice3.services.web_protocol.element import WebElement


class SwitchTo:
    def __init__(self, session):
        """
        :param pymobiledevice3.services.web_protocol.automation_session.AutomationSession session: Automation session.
        """
        self.session = session

    async def get_active_element(self) -> WebElement:
        """Returns the element with focus, or BODY if nothing has focus."""
        await self.session.wait_for_navigation_to_complete()
        elem = await self.session.evaluate_js_function(
            "function() { return document.activeElement; }", include_frame=False
        )
        return WebElement(self.session, elem)

    def get_alert(self) -> Alert:
        """Switches focus to an alert on the page."""
        return Alert(self.session)

    async def default_content(self):
        """Switch focus to the default frame."""
        await self.session.switch_to_browsing_context("")

    async def frame(self, frame_reference):
        """
        Switches focus to the specified frame, by index, name, or web element.
        :param frame_reference: The name of the window to switch to, an integer representing the index,
                                or a web element that is an (i)frame to switch to.
        """
        if isinstance(frame_reference, (int, WebElement)):
            frame = frame_reference
        elif isinstance(frame_reference, str):
            elem = await self.session.find_elements(By.ID, frame_reference)
            if elem is None:
                elem = await self.session.find_elements(By.NAME, frame_reference)
            frame = WebElement(self.session, elem)
        else:
            raise TypeError()

        await self.session.wait_for_navigation_to_complete()
        if isinstance(frame, int):
            await self.session.switch_to_frame(frame_ordinal=frame)
        else:
            await self.session.switch_to_frame(frame_handle=frame)

    async def new_window(self, type_=""):
        """Switches to a new top-level browsing context."""
        await self.session.switch_to_window(self.session.create_window(type_))

    async def parent_frame(self):
        """
        Switches focus to the parent context. If the current context is the top
        level browsing context, the context remains unchanged.
        """
        await self.session.wait_for_navigation_to_complete()
        await self.session.switch_to_browsing_context_frame(
            self.session.top_level_handle, self.session.current_parent_handle
        )
        await self.session.switch_to_browsing_context(self.session.current_parent_handle)

    async def window(self, window_name):
        """Switches focus to the specified window."""
        await self.session.switch_to_window(window_name)



import asyncio
import logging

from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command, print_json
from pymobiledevice3.services.diagnostics import DiagnosticsService

logger = logging.getLogger(__name__)

cli = InjectingTyper(
    name="battery",
    help="Battery options",
    no_args_is_help=True,
)


@cli.command("single")
@async_command
async def diagnostics_battery_single(service_provider: ServiceProviderDep) -> None:
    """get single snapshot of battery data"""
    raw_info = await DiagnosticsService(lockdown=service_provider).get_battery()
    print_json(raw_info)


@cli.command("monitor")
@async_command
async def diagnostics_battery_monitor(service_provider: ServiceProviderDep) -> None:
    """monitor battery usage"""
    diagnostics = DiagnosticsService(lockdown=service_provider)
    while True:
        raw_info = await diagnostics.get_battery()
        info = {
            "InstantAmperage": raw_info.get("InstantAmperage"),
            "Temperature": raw_info.get("Temperature"),
            "Voltage": raw_info.get("Voltage"),
            "IsCharging": raw_info.get("IsCharging"),
            "CurrentCapacity": raw_info.get("CurrentCapacity"),
        }
        logger.info(info)
        await asyncio.sleep(1)


@cli.command("wifi")
@async_command
async def diagnostics_wifi(service_provider: ServiceProviderDep) -> None:
    """Query WiFi info from IORegistry"""
    raw_info = await DiagnosticsService(lockdown=service_provider).get_wifi()
    print_json(raw_info)



import contextlib
import plistlib
from enum import Enum
from pathlib import Path
from typing import Any, Optional
from uuid import uuid4

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization.pkcs7 import PKCS7SignatureBuilder

from pymobiledevice3.exceptions import CloudConfigurationAlreadyPresentError, ConnectionTerminatedError, ProfileError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService

ERROR_CLOUD_CONFIGURATION_ALREADY_PRESENT = 14002
GLOBAL_HTTP_PROXY_UUID = "86a52338-52f7-4c09-b005-52baf3dc4882"
GLOBAL_RESTRICTIONS_UUID = "e22a0a66-08a8-43f5-9bbc-5279af35bb2b"


class Purpose(Enum):
    PostSetupInstallation = "PostSetupInstallation"


class MobileConfigService(LockdownService):
    SERVICE_NAME = "com.apple.mobile.MCInstall"
    RSD_SERVICE_NAME = "com.apple.mobile.MCInstall.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    async def hello(self) -> None:
        await self._send_recv({"RequestType": "HelloHostIdentifier"})

    async def flush(self) -> None:
        await self._send_recv({"RequestType": "Flush"})

    async def escalate(self, keybag_file: Path) -> None:
        """
        Authenticate with the device.

        :param keybag_file: Certificate file in PEM format, containing certificate and private key.
        :return: None
        """
        with open(keybag_file, "rb") as keybag_file:
            keybag_file = keybag_file.read()
        private_key = serialization.load_pem_private_key(keybag_file, password=None)
        cer = x509.load_pem_x509_certificate(keybag_file)
        public_key = cer.public_bytes(Encoding.DER)
        escalate_response = await self._send_recv({"RequestType": "Escalate", "SupervisorCertificate": public_key})
        signed_challenge = (
            PKCS7SignatureBuilder()
            .set_data(escalate_response["Challenge"])
            .add_signer(cer, private_key, hashes.SHA256())
            .sign(Encoding.DER, [])
        )
        await self._send_recv({"RequestType": "EscalateResponse", "SignedRequest": signed_challenge})
        await self._send_recv({"RequestType": "ProceedWithKeybagMigration"})

    async def get_stored_profile(self, purpose: Purpose = Purpose.PostSetupInstallation) -> dict:
        return await self._send_recv({"RequestType": "GetStoredProfile", "Purpose": purpose.value})

    async def store_profile(self, profile_data: bytes, purpose: Purpose = Purpose.PostSetupInstallation) -> None:
        await self._send_recv({"RequestType": "StoreProfile", "ProfileData": profile_data, "Purpose": purpose.value})

    async def get_cloud_configuration(self) -> dict:
        return (await self._send_recv({"RequestType": "GetCloudConfiguration"})).get("CloudConfiguration")

    async def set_cloud_configuration(self, cloud_configuration: dict) -> None:
        await self._send_recv({"RequestType": "SetCloudConfiguration", "CloudConfiguration": cloud_configuration})

    async def establish_provisional_enrollment(self, nonce: bytes) -> None:
        await self._send_recv({"RequestType": "EstablishProvisionalEnrollment", "Nonce": nonce})

    async def set_wifi_power_state(self, state: bool) -> None:
        await self._send_recv({"RequestType": "SetWiFiPowerState", "PowerState": state})

    async def erase_device(self, preserve_data_plan: bool, disallow_proximity_setup: bool) -> None:
        with contextlib.suppress(ConnectionTerminatedError):
            await self._send_recv({
                "RequestType": "EraseDevice",
                "PreserveDataPlan": preserve_data_plan,
                "DisallowProximitySetup": disallow_proximity_setup,
            })

    async def get_profile_list(self) -> dict:
        return await self._send_recv({"RequestType": "GetProfileList"})

    async def install_profile(self, payload: bytes) -> None:
        await self._send_recv({"RequestType": "InstallProfile", "Payload": payload})

    async def install_profile_silent(self, keybag_file: Path, profile: bytes) -> None:
        await self.escalate(keybag_file)
        await self._send_recv({"RequestType": "InstallProfileSilent", "Payload": profile})

    async def remove_profile(self, identifier: str) -> None:
        profiles = await self.get_profile_list()
        if not profiles:
            return
        if identifier not in profiles["ProfileMetadata"]:
            self.logger.info(f"Trying to remove not installed profile: {identifier}")
            return
        meta = profiles["ProfileMetadata"][identifier]
        data = plistlib.dumps({
            "PayloadType": "Configuration",
            "PayloadIdentifier": identifier,
            "PayloadUUID": meta["PayloadUUID"],
            "PayloadVersion": meta["PayloadVersion"],
        })
        await self._send_recv({"RequestType": "RemoveProfile", "ProfileIdentifier": data})

    async def _send_recv(self, request: dict) -> dict:
        response = await self.service.send_recv_plist(request)
        if response.get("Status", None) != "Acknowledged":
            error_chain = response.get("ErrorChain")
            if error_chain is not None:
                error_code = error_chain[0]["ErrorCode"]
                if error_code == ERROR_CLOUD_CONFIGURATION_ALREADY_PRESENT:
                    raise CloudConfigurationAlreadyPresentError()
            raise ProfileError(f"invalid response {response}")
        return response

    async def install_wifi_profile(
        self,
        encryption_type: str,
        ssid: str,
        password: str,
        auto_join: bool = True,
        captive_bypass: bool = False,
        disable_association_mac_randomization: bool = False,
        hidden_network: bool = False,
        is_hotspot: bool = False,
        keybag_file: Optional[Path] = None,
    ) -> None:
        payload_uuid = str(uuid4())
        await self.install_managed_profile(
            f"WiFi Profile For {ssid}",
            {
                "AutoJoin": auto_join,
                "CaptiveBypass": captive_bypass,
                "DisableAssociationMACRandomization": disable_association_mac_randomization,
                "EncryptionType": encryption_type,
                "HIDDEN_NETWORK": hidden_network,
                "IsHotspot": is_hotspot,
                "Password": password,
                "PayloadDescription": "Configures Wi-Fi settings",
                "PayloadDisplayName": "Wi-Fi",
                "PayloadIdentifier": f"com.apple.wifi.managed.{payload_uuid}",
                "PayloadType": "com.apple.wifi.managed",
                "PayloadUUID": payload_uuid,
                "PayloadVersion": 1,
                "ProxyType": "None",
                "SSID_STR": ssid,
            },
            keybag_file=keybag_file,
        )

    async def install_http_proxy(self, server: str, server_port: int, keybag_file: Optional[Path] = None) -> None:
        payload_uuid = str(uuid4())
        await self.install_managed_profile(
            f"HTTP Proxy for {server}:{server_port}",
            {
                "PayloadDescription": "Global HTTP Proxy",
                "PayloadDisplayName": "Global HTTP Proxy",
                "PayloadIdentifier": f"com.apple.proxy.http.global.{payload_uuid}",
                "PayloadType": "com.apple.proxy.http.global",
                "PayloadUUID": payload_uuid,
                "PayloadVersion": 1,
                "ProxyCaptiveLoginAllowed": False,
                "ProxyServer": server,
                "ProxyServerPort": server_port,
                "ProxyType": "Manual",
            },
            payload_uuid=GLOBAL_HTTP_PROXY_UUID,
            keybag_file=keybag_file,
        )

    async def remove_http_proxy(self) -> None:
        await self.remove_profile(GLOBAL_HTTP_PROXY_UUID)

    async def supervise(self, organization: str, keybag_file: Path) -> None:
        cer = x509.load_pem_x509_certificate(keybag_file.read_bytes())
        public_key = cer.public_bytes(Encoding.DER)
        await self.set_cloud_configuration({
            "AllowPairing": True,
            "CloudConfigurationUIComplete": True,
            "ConfigurationSource": 2,
            "ConfigurationWasApplied": True,
            "IsMDMUnremovable": False,
            "IsMandatory": True,
            "IsMultiUser": False,
            "IsSupervised": True,
            "OrganizationMagic": str(uuid4()),
            "OrganizationName": organization,
            "PostSetupProfileWasInstalled": True,
            "SkipSetup": [
                "Location",
                "Restore",
                "SIMSetup",
                "Android",
                "AppleID",
                "IntendedUser",
                "TOS",
                "Siri",
                "ScreenTime",
                "Diagnostics",
                "SoftwareUpdate",
                "Passcode",
                "Biometric",
                "Payment",
                "Zoom",
                "DisplayTone",
                "MessagingActivationUsingPhoneNumber",
                "HomeButtonSensitivity",
                "CloudStorage",
                "ScreenSaver",
                "TapToSetup",
                "Keyboard",
                "PreferredLanguage",
                "SpokenLanguage",
                "WatchMigration",
                "OnBoarding",
                "TVProviderSignIn",
                "TVHomeScreenSync",
                "Privacy",
                "TVRoom",
                "iMessageAndFaceTime",
                "AppStore",
                "Safety",
                "Multitasking",
                "ActionButton",
                "TermsOfAddress",
                "AccessibilityAppearance",
                "Welcome",
                "Appearance",
                "RestoreCompleted",
                "UpdateCompleted",
                "WiFi",
                "Display",
                "Tone",
                "LanguageAndLocale",
                "TouchID",
                "TrueToneDisplay",
                "FileVault",
                "iCloudStorage",
                "iCloudDiagnostics",
                "Registration",
                "DeviceToDeviceMigration",
                "UnlockWithWatch",
                "Accessibility",
                "All",
                "ExpressLanguage",
                "Language",
                "N/A",
                "Region",
                "Avatar",
                "DeviceProtection",
                "Key",
                "LockdownMode",
                "Wallpaper",
                "PrivacySubtitle",
                "SecuritySubtitle",
                "DataSubtitle",
                "AppleIDSubtitle",
                "AppearanceSubtitle",
                "PreferredLang",
                "OnboardingSubtitle",
                "AppleTVSubtitle",
                "Intelligence",
                "WebContentFiltering",
                "CameraButton",
                "AdditionalPrivacySettings",
                "EnableLockdownMode",
                "OSShowcase",
                "SafetyAndHandling",
                "Tips",
                "AgeBasedSafetySettings",
            ],
            "SupervisorHostCertificates": [public_key],
        })

    async def install_managed_profile(
        self,
        display_name: str,
        payload_content: dict[str, Any],
        payload_uuid: str = str(uuid4()),
        keybag_file: Optional[Path] = None,
    ) -> None:
        profile_data = plistlib.dumps({
            "PayloadContent": [payload_content],
            "PayloadDisplayName": display_name,
            "PayloadIdentifier": payload_uuid,
            "PayloadRemovalDisallowed": False,
            "PayloadType": "Configuration",
            "PayloadUUID": payload_uuid,
            "PayloadVersion": 1,
        })
        if keybag_file is not None:
            await self.install_profile_silent(keybag_file, profile_data)
        else:
            await self.install_profile(profile_data)

    async def install_restrictions_profile(
        self,
        enforced_software_update_delay: int = 0,
        payload_uuid: str = GLOBAL_RESTRICTIONS_UUID,
        keybag_file: Optional[Path] = None,
    ) -> None:
        await self.install_managed_profile(
            "Restrictions",
            {
                "PayloadDescription": "Configures restrictions",
                "PayloadDisplayName": "Restrictions",
                "PayloadIdentifier": f"com.apple.applicationaccess.{payload_uuid}",
                "PayloadType": "com.apple.applicationaccess",
                "PayloadUUID": payload_uuid,
                "PayloadVersion": 1,
                "allowActivityContinuation": True,
                "allowAddingGameCenterFriends": True,
                "allowAirPlayIncomingRequests": True,
                "allowAirPrint": True,
                "allowAirPrintCredentialsStorage": True,
                "allowAirPrintiBeaconDiscovery": True,
                "allowAppCellularDataModification": True,
                "allowAppClips": True,
                "allowAppInstallation": True,
                "allowAppRemoval": True,
                "allowApplePersonalizedAdvertising": True,
                "allowAssistant": True,
                "allowAssistantWhileLocked": True,
                "allowAutoCorrection": True,
                "allowAutoUnlock": True,
                "allowAutomaticAppDownloads": True,
                "allowBluetoothModification": True,
                "allowBookstore": True,
                "allowBookstoreErotica": True,
                "allowCamera": True,
                "allowCellularPlanModification": True,
                "allowChat": True,
                "allowCloudBackup": True,
                "allowCloudDocumentSync": True,
                "allowCloudPhotoLibrary": True,
                "allowContinuousPathKeyboard": True,
                "allowDefinitionLookup": True,
                "allowDeviceNameModification": True,
                "allowDeviceSleep": True,
                "allowDictation": True,
                "allowESIMModification": True,
                "allowEnablingRestrictions": True,
                "allowEnterpriseAppTrust": True,
                "allowEnterpriseBookBackup": True,
                "allowEnterpriseBookMetadataSync": True,
                "allowEraseContentAndSettings": True,
                "allowExplicitContent": True,
                "allowFilesNetworkDriveAccess": True,
                "allowFilesUSBDriveAccess": True,
                "allowFindMyDevice": True,
                "allowFindMyFriends": True,
                "allowFingerprintForUnlock": True,
                "allowFingerprintModification": True,
                "allowGameCenter": True,
                "allowGlobalBackgroundFetchWhenRoaming": True,
                "allowInAppPurchases": True,
                "allowKeyboardShortcuts": True,
                "allowManagedAppsCloudSync": True,
                "allowMultiplayerGaming": True,
                "allowMusicService": True,
                "allowNews": True,
                "allowNotificationsModification": True,
                "allowOpenFromManagedToUnmanaged": True,
                "allowOpenFromUnmanagedToManaged": True,
                "allowPairedWatch": True,
                "allowPassbookWhileLocked": True,
                "allowPasscodeModification": True,
                "allowPasswordAutoFill": True,
                "allowPasswordProximityRequests": True,
                "allowPasswordSharing": True,
                "allowPersonalHotspotModification": True,
                "allowPhotoStream": True,
                "allowPredictiveKeyboard": True,
                "allowProximitySetupToNewDevice": True,
                "allowRadioService": True,
                "allowRemoteAppPairing": True,
                "allowRemoteScreenObservation": True,
                "allowSafari": True,
                "allowScreenShot": True,
                "allowSharedStream": True,
                "allowSpellCheck": True,
                "allowSpotlightInternetResults": True,
                "allowSystemAppRemoval": True,
                "allowUIAppInstallation": True,
                "allowUIConfigurationProfileInstallation": True,
                "allowUSBRestrictedMode": True,
                "allowUnpairedExternalBootToRecovery": False,
                "allowUntrustedTLSPrompt": True,
                "allowVPNCreation": True,
                "allowVideoConferencing": True,
                "allowVoiceDialing": True,
                "allowWallpaperModification": True,
                "allowiTunes": True,
                "enforcedSoftwareUpdateDelay": enforced_software_update_delay,
                "forceAirDropUnmanaged": False,
                "forceAirPrintTrustedTLSRequirement": False,
                "forceAssistantProfanityFilter": False,
                "forceAuthenticationBeforeAutoFill": False,
                "forceAutomaticDateAndTime": False,
                "forceClassroomAutomaticallyJoinClasses": False,
                "forceClassroomRequestPermissionToLeaveClasses": False,
                "forceClassroomUnpromptedAppAndDeviceLock": False,
                "forceClassroomUnpromptedScreenObservation": False,
                "forceDelayedSoftwareUpdates": True,
                "forceEncryptedBackup": False,
                "forceITunesStorePasswordEntry": False,
                "forceLimitAdTracking": False,
                "forceWatchWristDetection": False,
                "forceWiFiPowerOn": False,
                "forceWiFiWhitelisting": False,
                "ratingApps": 1000,
                "ratingMovies": 1000,
                "ratingRegion": "us",
                "ratingTVShows": 1000,
                "safariAcceptCookies": 2.0,
                "safariAllowAutoFill": True,
                "safariAllowJavaScript": True,
                "safariAllowPopups": True,
                "safariForceFraudWarning": False,
            },
            payload_uuid=payload_uuid,
            keybag_file=keybag_file,
        )



from enum import IntEnum
from typing import Optional

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.service_connection import build_plist
from pymobiledevice3.services.lockdown_service import LockdownService


class InterfaceOrientation(IntEnum):
    PORTRAIT = 1
    PORTRAIT_UPSIDE_DOWN = 2
    LANDSCAPE = 3
    LANDSCAPE_HOME_TO_LEFT = 4


class SpringBoardServicesService(LockdownService):
    RSD_SERVICE_NAME = "com.apple.springboardservices.shim.remote"
    SERVICE_NAME = "com.apple.springboardservices"

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    async def get_icon_state(self, format_version: str = "2") -> list:
        cmd = {"command": "getIconState"}
        if format_version:
            cmd["formatVersion"] = format_version
        return await self.service.send_recv_plist(cmd)

    async def set_icon_state(self, newstate: Optional[list] = None) -> None:
        if newstate is None:
            newstate = {}
        await self.service.send_recv_prefixed(build_plist({"command": "setIconState", "iconState": newstate}))

    async def get_icon_pngdata(self, bundle_id: str) -> bytes:
        return (await self.service.send_recv_plist({"command": "getIconPNGData", "bundleId": bundle_id})).get("pngData")

    async def get_interface_orientation(self) -> InterfaceOrientation:
        res = await self.service.send_recv_plist({"command": "getInterfaceOrientation"})
        return InterfaceOrientation(res.get("interfaceOrientation"))

    async def get_wallpaper_pngdata(self) -> bytes:
        return (await self.service.send_recv_plist({"command": "getHomeScreenWallpaperPNGData"})).get("pngData")

    async def get_homescreen_icon_metrics(self) -> dict[str, float]:
        return await self.service.send_recv_plist({"command": "getHomeScreenIconMetrics"})

    async def get_wallpaper_info(self, wallpaper_name: str) -> dict:
        return await self.service.send_recv_plist({"command": "getWallpaperInfo", "wallpaperName": wallpaper_name})

    async def reload_icon_state(self) -> None:
        await self.set_icon_state(await self.get_icon_state())

    async def get_wallpaper_preview_image(self, wallpaper_name: str) -> bytes:
        return (
            await self.service.send_recv_plist({"command": "getWallpaperPreviewImage", "wallpaperName": wallpaper_name})
        )["pngData"]
